import logging

import numpy as np
import scipy.sparse as sparse
from scipy.fft import dct, idct
from scipy.special import jv

from aspire.basis import FBBasisMixin, SteerableBasis2D
from aspire.basis.basis_utils import besselj_zeros
from aspire.basis.fle_2d_utils import (
    barycentric_interp_sparse,
    precomp_transform_complex_to_real,
    transform_complex_to_real,
)
from aspire.nufft import anufft, nufft
from aspire.numeric import fft
from aspire.utils import complex_type, grid_2d

logger = logging.getLogger(__name__)


class FLEBasis2D(SteerableBasis2D, FBBasisMixin):
    """
    Define a derived class for Fast Fourier Bessel 2D expansion using interpolation
        from Chebyshev nodes.
    The algorithms used are described in the following publication:
    N. F. Marshall, O. Mickelin, A. Singer, Fast Expansion into Harmonics on the Disk:
        A Steerable Basis with Fast Radial Convolution. (submitted)

    https://arxiv.org/pdf/2207.13674.pdf
    """

    def __init__(
        self, size, bandlimit=None, epsilon=1e-10, dtype=np.float32, match_fb=True
    ):
        """
        :param size: The size of the vectors for which to define the FLE basis.
            Currently only square images are supported.
        :param bandlimit: Maximum frequency band for computing basis functions. Defaults to the
            resolution of the basis.
        :param epsilon: Relative precision between FLE fast method and dense matrix multiplication.
        :param dtype: Datatype of images and coefficients represented.
        :param match_fb: This flag constructs basis functions
            identical to `FBBasis2D`. The initial heuristic for the
            number of basis functions, based on the image size, will
            be set to that of `FBBasis2D`, and the FLE frequency
            thresholding procedure to reduce the number of functions
            will not be carried out. This means the number of basis
            functions for a given image size will be identical across
            the two bases.
        """
        if isinstance(size, int):
            size = (size, size)
        ndim = len(size)
        assert ndim == 2, "Only two-dimensional basis functions are supported."
        assert len(set(size)) == 1, "Only square domains are supported"

        self.bandlimit = bandlimit
        self.epsilon = epsilon
        self.match_fb = match_fb
        self.dtype = dtype
        super().__init__(size, ell_max=None, dtype=self.dtype)

    def _build(self):
        """
        Build the internal data structure for the FLEBasis2D class.
        """

        # bandlimit set to basis size by default
        if not self.bandlimit:
            self.bandlimit = self.nres

        # compute number of k's for each ell
        self._calc_k_max()

        if self.match_fb:
            # Use FB2D and FFB2D heuristic for computing max basis functions
            self.max_basis_functions = self.k_max[0] + sum(2 * self.k_max[1:])
        else:
            # Regular Fourier-Bessel bandlimit (equivalent to pi*R**2)
            # Final self.count will be < self.max_basis_functions
            # See self._threshold_basis_functions()
            self.max_basis_functions = int(self.nres**2 * np.pi / 4)

        self._compute_maxitr_and_numsparse()

        self._compute_cartesian_gridpoints()

        self._precomp()

        # Steerable basis indices
        self._build_indices()

    def _build_indices(self):
        # FLE internal indices
        self._fle_angular_indices = np.abs(self._ells)
        self._fle_radial_indices = self._ks - 1
        # Negate all signs from FLE implementation
        self._fle_signs_indices = -np.sign(self._ells)
        # Use the FB2D ells sign convention of `1` for `ell=0`
        self._fle_signs_indices[self._ells == 0] = 1

        # basis function ordering (used during evaluate_t output)
        self._fle_to_fb_indices = np.lexsort(
            (
                self._fle_radial_indices,
                # Reverse sign sorting order so +1 first,
                #   match `sgns = (1,) if ell == 0 else (1, -1)` from fb_2d.py
                -self._fle_signs_indices,
                self._fle_angular_indices,
            )
        )
        # store the reverse mapping (used during evaluate input)
        self._fb_to_fle_indices = np.argsort(self._fle_to_fb_indices)

        # User facing indices, should follow FB ordering.
        self.angular_indices = self._fle_angular_indices[self._fle_to_fb_indices]
        self.radial_indices = self._fle_radial_indices[self._fle_to_fb_indices]
        # Note we negate the FLE signs?
        self.signs_indices = self._fle_signs_indices[self._fle_to_fb_indices]

    def indices(self):
        """
        Return the precomputed indices for each basis function.
        """
        return {
            "ells": self.angular_indices,
            "ks": self.radial_indices,
            "sgns": self.signs_indices,
        }

    def _precomp(self):
        """
        Precompute the basis functions and other objects used in the evaluation of
            coefficients.
        """

        # Find bessel functions zeros (the eigenvalues of the Laplacian on
        # the disk) and generate the FLE Basis functions
        self._lap_eig_disk()

        # Some important constants
        self.smallest_lambda = np.min(self.bessel_zeros)
        self.greatest_lambda = np.max(self.bessel_zeros)
        self.max_ell = np.max(np.abs(self._ells))
        self.h = 1 / (self.nres / 2)

        # give each ell a positive index increasing first in |ell|
        # then in sign, e.g. 0->1, -1->2, 1->3, -2->4, 2->5, etc.
        self._ells_p = 2 * np.abs(self._ells) - (self._ells < 0)
        self.ell_p_max = np.max(self._ells_p)
        # idx_list[k] contains the indices j of ells_p where ells_p[j] = k
        idx_list = [[] for i in range(self.ell_p_max + 1)]
        for i in range(self.count):
            ellp = self._ells_p[i]
            idx_list[ellp].append(i)
        self.idx_list = idx_list

        # real <-> complex
        self.c2r = precomp_transform_complex_to_real(self._ells)
        self.r2c = sparse.csr_matrix(self.c2r.transpose().conj())

        # create an ordered list of the original ell values
        # used in step2 (in both directions)
        self.nus = np.zeros(1 + 2 * self.max_ell, dtype=int)
        self.nus[0] = 0
        for i in range(1, self.max_ell + 1):
            self.nus[2 * i - 1] = -i
            self.nus[2 * i] = i
        self.c2r_nus = precomp_transform_complex_to_real(self.nus)
        self.r2c_nus = sparse.csr_matrix(self.c2r_nus.transpose().conj())

        # radial and angular nodes for NUFFT
        self._compute_nufft_points()
        self.num_interp = self.num_radial_nodes
        if self.numsparse > 0:
            self.num_interp = 2 * self.num_radial_nodes

        self._build_interpolation_matrix()

    def _compute_maxitr_and_numsparse(self):
        """
        Uses heuristics from paper to assign self.maxitr and self.numsparse.
        """
        # maxitr: maximum number of iterations for numerically solving linear
        # system in self.evaluate()
        # numsparse: parameter used to create sparse Chebyshev interpolation matrix
        # see self._build_interpolation_matrix()

        if self.epsilon >= 1e-4:
            numsparse = 8
            maxitr = 1 + int(np.log2(self.nres)) // 2
        elif self.epsilon >= 1e-7:
            numsparse = 16
            maxitr = 1 + int(np.log2(self.nres))
        elif self.epsilon >= 1e-10:
            numsparse = 22
            maxitr = 1 + int(2 * np.log2(self.nres))
        else:
            # epsilon < 1e-10
            numsparse = 32
            maxitr = 1 + int(3 * np.log2(self.nres))

        self.maxitr = maxitr
        self.numsparse = numsparse

    def _compute_cartesian_gridpoints(self):
        """
        Creates meshgrids based on basis size.
        """
        if self.match_fb:
            # creates correct odd-resolution grid
            # matching other FB classes
            grid = grid_2d(self.nres, dtype=self.dtype)
            self.xs = grid["x"]
            self.ys = grid["y"]
            self.rs = grid["r"]
        else:
            # original implementation
            R = self.nres // 2
            x = np.arange(-R, R + self.nres % 2)
            xs, ys = np.meshgrid(x, x)
            # Note, the original original grids were xs/R, R=nres//2.
            self.xs, self.ys = xs / (self.nres / 2), ys / (self.nres / 2)
            self.rs = np.sqrt(self.xs**2 + self.ys**2)
        self.radial_mask = self.rs > 1 + 1e-13

    def _compute_nufft_points(self):
        """
        Computes gridpoints for the non-uniform FFT.
        """

        # Number of radial nodes
        # (Lemma 4.1)
        # compute max {2.4 * self.nres , Log2 ( 1 / epsilon) }
        Q = int(np.ceil(2.4 * self.nres))
        num_radial_nodes = Q
        tmp = 1 / (np.sqrt(np.pi))
        for q in range(1, Q + 1):
            tmp = tmp / q * (np.sqrt(np.pi) * self.nres / 4)
            if tmp <= self.epsilon:
                num_radial_nodes = int(max(q, np.log2(1 / self.epsilon)))
                break
        self.num_radial_nodes = max(
            num_radial_nodes, int(np.ceil(np.log2(1 / self.epsilon)))
        )

        # Number of angular nodes
        # (Lemma 4.2)
        # compute max {7.08 * self.nres, Log2(1/epsilon) + Log2(self.nres**2) }

        S = int(max(7.08 * self.nres, -np.log2(self.epsilon) + 2 * np.log2(self.nres)))
        num_angular_nodes = S
        for s in range(int(self.greatest_lambda + self.ell_p_max) + 1, S + 1):
            tmp = self.nres**2 * ((self.greatest_lambda + self.ell_p_max) / s) ** s
            if tmp <= self.epsilon:
                num_angular_nodes = int(max(int(s), np.log2(1 / self.epsilon)))
                break

        # must be even
        if num_angular_nodes % 2 == 1:
            num_angular_nodes += 1

        self.num_angular_nodes = num_angular_nodes

        # create gridpoints
        nodes = 1 - (2 * np.arange(self.num_radial_nodes, dtype=self.dtype) + 1) / (
            2 * self.num_radial_nodes
        )
        nodes = (np.cos(np.pi * nodes) + 1) / 2
        nodes = (
            self.greatest_lambda - self.smallest_lambda
        ) * nodes + self.smallest_lambda
        nodes = nodes.reshape(self.num_radial_nodes, 1)

        radius = self.nres / 2
        h = 1 / radius

        phi = (
            2
            * np.pi
            * np.arange(self.num_angular_nodes // 2, dtype=self.dtype)
            / self.num_angular_nodes
        )
        x = np.cos(phi).reshape(1, self.num_angular_nodes // 2)
        y = np.sin(phi).reshape(1, self.num_angular_nodes // 2)
        x = x * nodes * h
        y = y * nodes * h
        self.grid_x = x.flatten()
        self.grid_y = y.flatten()

    def _build_interpolation_matrix(self):
        """
        Create the matrix used in the third step of evaluate_t() and the first step of evaluate()
        for barycentric interpolation from Chebyshev nodes.
        """
        A3 = [None] * (self.ell_p_max + 1)
        A3_T = [None] * (self.ell_p_max + 1)
        # known points from which to interpolate Beta values to desired points
        known_points = np.cos(
            np.pi * (1 - (2 * np.arange(self.num_interp) + 1) / (2 * self.num_interp))
        )
        for i in range(self.ell_p_max + 1):
            # target points to evaluate Betas
            target_points = (
                2
                * (self.bessel_zeros[self.idx_list[i]] - self.smallest_lambda)
                / (self.greatest_lambda - self.smallest_lambda)
                - 1
            )

            A3[i], A3_T[i] = barycentric_interp_sparse(
                target_points, known_points, self.numsparse
            )
        self.A3 = A3
        self.A3_T = A3_T

    def _lap_eig_disk(self):
        """
        Compute the eigenvalues of the Laplacian operator on a disk with Dirichlet boundary conditions.
        """
        # max number of Bessel function orders being considered
        max_ell = int(3 * np.sqrt(self.max_basis_functions))
        # max number of zeros per Bessel function (number of frequencies per bessel)
        max_k = int(2 * np.sqrt(self.max_basis_functions))

        # preallocate containers for roots
        # 0 frequency plus pos and negative frequencies for each bessel function
        # num functions per frequency
        num_ells = 1 + 2 * max_ell
        self._ells = np.zeros((num_ells, max_k), dtype=int)
        self._ks = np.zeros((num_ells, max_k), dtype=int)
        self.bessel_zeros = np.ones((num_ells, max_k), dtype=np.float64) * np.Inf

        # keep track of which order Bessel function we're on
        self._ells[0, :] = 0
        # bessel_roots[0, m] is the m'th zero of J_0
        self.bessel_zeros[0, :] = besselj_zeros(0, max_k)
        # table of values of which zero of J_0 we are finding
        self._ks[0, :] = np.arange(max_k) + 1

        # add roots of J_ell for ell>0 twice with +k and -k (frequencies)
        # iterate over Bessel function order
        for ell in range(1, max_ell + 1):
            self._ells[2 * ell - 1, :] = -ell
            self._ks[2 * ell - 1, :] = np.arange(max_k) + 1

            self.bessel_zeros[2 * ell - 1, :max_k] = besselj_zeros(ell, max_k)

            self._ells[2 * ell, :] = ell
            self._ks[2 * ell, :] = self._ks[2 * ell - 1, :]
            self.bessel_zeros[2 * ell, :] = self.bessel_zeros[2 * ell - 1, :]

        # Reshape the arrays and order by the size of the Bessel function zeros
        self._flatten_and_sort_bessel_zeros()

        # Apply threshold criterion to throw out some basis functions
        # Grab final number of basis functions for this Basis
        self.count = self._threshold_basis_functions()

        self._create_basis_functions()

    def _flatten_and_sort_bessel_zeros(self):
        """
        Reshapes arrays self._ells, self._ks, and self.bessel_zeros
        """
        # flatten list of zeros, ells and ks:
        self._ells = self._ells.flatten()
        self._ks = self._ks.flatten()
        self.bessel_zeros = self.bessel_zeros.flatten()

        idx = np.argsort(self.bessel_zeros)
        self._ells = self._ells[idx]
        self._ks = self._ks[idx]
        self.bessel_zeros = self.bessel_zeros[idx]

        # sort complex conjugate pairs: -ell first, +ell second
        idx = np.arange(self.max_basis_functions + 1)
        for i in range(self.max_basis_functions + 1):
            if self._ells[i] >= 0:
                continue
            if np.abs(self.bessel_zeros[i] - self.bessel_zeros[i + 1]) < 1e-14:
                continue
            idx[i - 1] = i
            idx[i] = i - 1

        self._ells = self._ells[idx]
        self._ks = self._ks[idx]
        self.bessel_zeros = self.bessel_zeros[idx]

    def _threshold_basis_functions(self):
        """
        Implements the bandlimit threshold which caps the number of basis functions
        that are actually required.
        :return: The final overall number of basis functions to be used.
        """
        # Maximum bandlimit
        # (Section 4.1)
        # Can remove frequencies above this threshold based on the fact that
        # there should not be more basis functions than pixels contained in the
        # unit disk inscribed on the image
        _final_num_basis_functions = self.max_basis_functions

        # implement FLE thresholding unless we want to match count of other FB bases
        if not self.match_fb:
            for _ in range(len(self.bessel_zeros)):
                if (
                    self.bessel_zeros[_final_num_basis_functions] / (np.pi)
                    >= (self.bandlimit - 1) // 2
                ):
                    _final_num_basis_functions -= 1

        # potentially subtract one to keep complex conjugate pairs
        if self._ells[_final_num_basis_functions - 1] < 0:
            _final_num_basis_functions -= 1

        # discard zeros above the threshold
        self._ells = self._ells[:_final_num_basis_functions]
        self._ks = self._ks[:_final_num_basis_functions]
        self.bessel_zeros = self.bessel_zeros[:_final_num_basis_functions]

        return _final_num_basis_functions

    def _create_basis_functions(self):
        """
        Generate the actual basis functions as Python lambda operators
        """
        norm_constants = np.zeros(self.count)
        basis_functions = [None] * self.count
        for i in range(self.count):
            # parameters defining the basis function: bessel order and which bessel root
            ell = self._ells[i]
            bessel_zero = self.bessel_zeros[i]

            # compute normalization constant
            # see Eq. 6
            c = 1 / np.sqrt(np.pi * jv(ell + 1, bessel_zero) ** 2)
            # create function
            # See Eq. 1
            if ell == 0:
                basis_functions[i] = (
                    lambda r, t, c=c, ell=ell, bessel_zero=bessel_zero: c
                    * jv(ell, bessel_zero * r)
                    * (r <= 1)
                )
            else:
                basis_functions[i] = (
                    lambda r, t, c=c, ell=ell, bessel_zero=bessel_zero: c
                    * jv(ell, bessel_zero * r)
                    * np.exp(1j * ell * t)
                    * (-1) ** np.abs(ell)
                    * (r <= 1)
                )

            norm_constants[i] = c

        self.norm_constants = norm_constants
        self.basis_functions = basis_functions

    def _evaluate(self, coeffs):
        """
        Evaluates FLE coefficients and return in standard 2D Cartesian coordinates.

        :param v: A coefficient vector (or an array of coefficient vectors) to
            be evaluated. The last dimension must be equal to `self.count`
        :return: An Image object containing the corresponding images.
        """
        # convert from FB order
        coeffs = coeffs[..., self._fb_to_fle_indices]

        # See Remark 3.3 and Section 3.4
        betas = self._step3(coeffs)
        z = self._step2(betas)
        im = self._step1(z)
        return im.astype(self.dtype)

    def _evaluate_t(self, imgs):
        """
        Evaluate 2D Cartesian image(s) and return the corresponding FLE coefficients.

        :param imgs: An Image object containing square images of size `self.nres`.
        :return: A NumPy array of size `(num_images, self.count)` containing the FLE
            coefficients.
        """
        # See Section 3.5
        imgs = imgs.copy()
        imgs[:, self.radial_mask] = 0
        z = self._step1_t(imgs)
        b = self._step2_t(z)
        coeffs = self._step3_t(b)

        # return in FB order
        coeffs = coeffs[..., self._fle_to_fb_indices]
        return coeffs.astype(self.coefficient_dtype, copy=False)

    def _step1_t(self, im):
        """
        Step 1 of the adjoint transformation (images to coefficients).
        Calculates the NUFFT of the image on gridpoints `self.grid_x` and `self.grid_y`.
        """
        im = im.reshape(-1, self.nres, self.nres).astype(complex_type(self.dtype))
        num_img = im.shape[0]
        z = np.zeros(
            (num_img, self.num_radial_nodes, self.num_angular_nodes),
            dtype=complex_type(self.dtype),
        )
        _z = (
            nufft(im, np.stack((self.grid_x, self.grid_y)), epsilon=self.epsilon)
            * self.h**2
        )
        _z = _z.reshape(num_img, self.num_radial_nodes, self.num_angular_nodes // 2)
        z[:, :, : self.num_angular_nodes // 2] = _z
        z[:, :, self.num_angular_nodes // 2 :] = np.conj(_z)
        return z

    def _step2_t(self, z):
        """
        Step 2 of the adjoint transformation (images to coefficients).
        Computes values of the analytic functions Beta_n at the Chebyshev nodes.
        See Lemma 2.2.
        """
        num_img = z.shape[0]
        # Compute FFT along angular nodes
        betas = fft.fft(z, axis=2) / self.num_angular_nodes
        betas = betas[:, :, self.nus]
        betas = np.conj(betas)
        betas = np.swapaxes(betas, 0, 2)
        betas = betas.reshape(-1, self.num_radial_nodes * num_img)
        betas = self.c2r_nus @ betas
        betas = betas.reshape(-1, self.num_radial_nodes, num_img)
        betas = np.real(np.swapaxes(betas, 0, 2))
        return betas

    def _step3_t(self, betas):
        """
        Step 3 of the adjoint transformation (images to coefficients).
        Uses barycenteric interpolation to compute the values of the Betas
        at the Bessel roots to arrive at the Fourier-Bessel coefficients.
        """
        num_img = betas.shape[0]
        if self.num_interp > self.num_radial_nodes:
            betas = dct(betas, axis=1, type=2) / (2 * self.num_radial_nodes)
            zeros = np.zeros(betas.shape)
            betas = np.concatenate((betas, zeros), axis=1)
            betas = idct(betas, axis=1, type=2) * 2 * betas.shape[1]
        betas = np.moveaxis(betas, 0, -1)

        coeffs = np.zeros((self.count, num_img), dtype=np.float64)
        for i in range(self.ell_p_max + 1):
            coeffs[self.idx_list[i]] = self.A3[i] @ betas[:, i, :]
        coeffs = coeffs.T

        return coeffs * self.norm_constants / self.h

    def _step3(self, coeffs):
        """
        Adjoint of _step3_t and Step 1 of the forward transformation (coefficients
            to images).
        Uses barycenteric interpolation in reverse to compute values of Betas
            at Chebyshev nodes, given an array of FLE coefficients.
        """
        coeffs = coeffs.copy().reshape(-1, self.count)
        num_img = coeffs.shape[0]
        coeffs *= self.h * self.norm_constants
        coeffs = coeffs.T

        out = np.zeros(
            (self.num_interp, 2 * self.max_ell + 1, num_img),
            dtype=np.float64,
        )
        for i in range(self.ell_p_max + 1):
            out[:, i, :] = self.A3_T[i] @ coeffs[self.idx_list[i]]
        out = np.moveaxis(out, -1, 0)
        if self.num_interp > self.num_radial_nodes:
            out = dct(out, axis=1, type=2)
            out = out[:, : self.num_radial_nodes, :]
            out = idct(out, axis=1, type=2)

        return out

    def _step2(self, betas):
        """
        Adjoint of _step2_t and Step 2 of the forward transformation (coefficients
            to images).
        Uses the IFFT to convert Beta values into Fourier-space images.
        """
        num_img = betas.shape[0]
        tmp = np.zeros(
            (num_img, self.num_radial_nodes, self.num_angular_nodes),
            dtype=np.complex128,
        )

        betas = np.swapaxes(betas, 0, 2)
        betas = betas.reshape(-1, self.num_radial_nodes * num_img)
        betas = self.r2c_nus @ betas
        betas = betas.reshape(-1, self.num_radial_nodes, num_img)
        betas = np.swapaxes(betas, 0, 2)

        tmp[:, :, self.nus] = np.conj(betas)
        z = fft.ifft(tmp, axis=2)

        return z

    def _step1(self, z):
        """
        Adjoint of _step1_t and final step of the forward transformation (coefficients
            to images).
        Performs the NUFFT on Fourier-space images to compute real-space images.
        """
        num_img = z.shape[0]
        z = z[:, :, : self.num_angular_nodes // 2].reshape(num_img, -1)
        im = anufft(
            z.astype(complex_type(self.dtype)),
            np.stack((self.grid_x, self.grid_y)),
            (self.nres, self.nres),
            epsilon=self.epsilon,
        )
        im = im + np.conj(im)
        im = np.real(im)
        im = im.reshape(num_img, self.nres, self.nres)
        im[:, self.radial_mask] = 0

        return im

    def _create_dense_matrix(self):
        """
        Directly computes the transformation matrix from Cartesian coordinates to
        FLE coordinates without any shortcuts.
        :return: A NumPy array of size `(self.nres**2, self.count)` containing the matrix
            entries.
        """
        # See Eqns. 3 and 4, Section 1.2
        ts = np.arctan2(self.ys, self.xs)

        B = np.zeros((self.nres, self.nres, self.count), dtype=np.complex128)
        for i in range(self.count):
            B[:, :, i] = self.basis_functions[i](self.rs, ts) * self.h
        B = B.reshape(self.nres**2, self.count)
        B = transform_complex_to_real(B, self._ells)
        B = B.reshape(self.nres**2, self.count)
        B = B[..., self._fle_to_fb_indices]

        return B

    def lowpass(self, coeffs, bandlimit):
        """
        Apply a low-pass filter to FLE coefficients `coeffs` with threshold `bandlimit`.
        :param coeffs: A NumPy array of FLE coefficients, of shape (num_images, self.count)
        :param bandlimit: Integer bandlimit (max frequency).
        :return: Band-limited coefficient array.
        """
        if len(coeffs.shape) == 1:
            coeffs = coeffs.reshape((1, coeffs.shape[0]))
        assert (
            len(coeffs.shape) == 2
        ), "Input a stack of coefficients of dimension (num_images, self.count)."
        assert (
            coeffs.shape[1] == self.count
        ), "Number of coefficients must match self.count."

        k = self.count - 1
        for _ in range(self.count):
            if self.bessel_zeros[k] / (np.pi) > (bandlimit - 1) // 2:
                k = k - 1
        coeffs[:, k + 1 :] = 0

        return coeffs

    def radial_convolve(self, coeffs, radial_img):
        """
        Convolve a stack of FLE coefficients with a 2D radial function.
        :param coeffs: A NumPy array of FLE coefficients of size (num_images, self.count).
        :param radial_img: A 2D NumPy array of size (self.nres, self.nres).
        :return: Convolved FLE coefficients.
        """
        num_img = coeffs.shape[0]
        coeffs_conv = np.zeros(coeffs.shape)

        # Convert to internal FLE indices ordering
        coeffs = coeffs[..., self._fb_to_fle_indices]

        for k in range(num_img):
            _coeffs = coeffs[k, :]
            z = self._step1_t(radial_img)
            b = self._step2_t(z)
            weights = self._radial_convolve_weights(b)
            b = weights / (self.h**2)
            b = b.reshape(self.count)
            coeffs_conv[k, :] = np.real(self.c2r @ (b * (self.r2c @ _coeffs).flatten()))

        # Convert from internal FLE ordering to FB convention
        coeffs_conv = coeffs_conv[..., self._fle_to_fb_indices]

        return coeffs_conv

    def _radial_convolve_weights(self, b):
        """
        Helper function for step 3 of convolving with a radial function.
        """
        b = np.squeeze(b)
        b = np.array(b)
        if self.num_interp > self.num_radial_nodes:
            b = dct(b, axis=0, type=2) / (2 * self.num_radial_nodes)
            bz = np.zeros(b.shape)
            b = np.concatenate((b, bz), axis=0)
            b = idct(b, axis=0, type=2) * 2 * b.shape[0]
        a = np.zeros(self.count, dtype=np.float64)
        y = [None] * (self.ell_p_max + 1)
        for i in range(self.ell_p_max + 1):
            y[i] = (self.A3[i] @ b[:, 0]).flatten()
        for i in range(self.ell_p_max + 1):
            a[self.idx_list[i]] = y[i]

        return a.flatten()



"""
Implements operations for diagonal matrices as used by ASPIRE.
"""

import numpy as np

from .blk_diag_matrix import BlkDiagMatrix, is_scalar_type


class DiagMatrix:
    """
    Implements operations for diagonal matrices as used by ASPIRE.

    Currently `DiagMatrix` is implemented as the equivalent of square
    matrices.  In the future it can be extended to support
    applications with rectangular shapes.
    """

    # Developers' Note:
    # All instances of this class should have priority over ndarray ops
    #   because we implement them here ourselves.
    # This is a more np current implementation of __array_priority__
    #   operator precedence schedule.
    # Mainly this effects rmul, eg:
    # np_ary @ diag_mat
    __array_ufunc__ = None

    def __init__(self, data, dtype=None):
        """
        Instantiate a `DiagMatrix` with Numpy `data` shaped (...., self.count),
        where `self.count` is the length of one diagonal vector.

        Slower axes (if present) are taken to be stack axes.

        :param data: Diagonal matrix entries.
        :param dtype: Datatype. Default of `None` will attempt
            passthrough of `data.dtype`.  When explicitly provided, will
            attempt casting `data` as needed.
        :return: `DiagMatrix` instance.
        """

        # Assign the datatype.
        if dtype is None:
            dtype = data.dtype
        self.dtype = np.dtype(dtype)

        # Assign the `data`
        self._data = data.astype(self.dtype, copy=False)

        # Assign shapes from `data`
        self.count = self._data.shape[-1]
        self.stack_shape = self._data.shape[:-1]
        # Total number of stack elements
        self.size = np.prod(self.stack_shape)

        # Numpy interop
        # https://numpy.org/devdocs/user/basics.interoperability.html#the-array-interface-protocol
        self.__array_interface__ = self.asnumpy().__array_interface__
        self.__array__ = self.asnumpy()

    def stack_reshape(self, *args):
        """
        Reshape the stack axis.

        :*args: Integer(s) or tuple describing the intended shape.

        :returns: `DiagMatrix` instance.
        """

        # If we're passed a tuple, use that
        if len(args) == 1 and isinstance(args[0], tuple):
            shape = args[0]
        else:
            # Otherwise use the variadic args
            shape = args

        # Sanity check the size
        if shape != (-1,) and np.prod(shape) != self.size:
            raise ValueError(
                f"Number of images {self.size} cannot be reshaped to {shape}."
            )

        return DiagMatrix(self._data.reshape(*shape, self._data.shape[-1]))

    def asnumpy(self):
        """
        Return data as Numpy array.

        Note this is a read-only view.

        :return: Numpy array of `self.dtype`.
        """

        view = self._data.view()
        view.flags.writeable = False
        return view

    def __repr__(self):
        """
        String represention describing instance.
        """
        return "DiagMatrix({}, {})".format(repr(self._data), repr(self.dtype))

    def copy(self):
        """
        Returns new `DiagMatrix` which is a copy of `self`.

        :return `DiagMatrix` like self
        """

        return DiagMatrix(self._data.copy())

    def __getitem__(self, key):
        """
        Convenience wrapper, getter on self._data.
        """

        return self._data[key]

    def __len__(self):
        """
        Convenience function for getting length of slowest stack axis.
        """
        return self.stack_shape[0]

    def __check_compatible(self, other):
        """
        Sanity check two `DiagMatrix` instances are compatible
        """

        self.__check_size_compatible(other)
        self.__check_dtype_compatible(other)

    def __check_size_compatible(self, other):
        """
        Sanity check two `DiagMatrix` instances are compatible in size
        for addition operators. (Same size)

        :param other: The `DiagMatrix` to compare with self.
        """

        if self.count != other.count:
            raise RuntimeError(
                "DiagMatrix instances are "
                f"not same dimension {self.count} != {other.count}"
            )
        try:
            _ = np.broadcast_shapes(self.stack_shape, other.stack_shape)
        except ValueError:
            raise ValueError(
                f"Attempting to broadcast incompatible shapes: {self.stack_shape} with {other.stack_shape}."
            )

    def __check_dtype_compatible(self, other):
        """
        Sanity check two `DiagMatrix` instances are compatible in dtype.

        :param other: The `DiagMatrix` to compare with self.
        """

        if self.dtype != other.dtype:
            raise RuntimeError(
                "DiagMatrix received different types,"
                f"self: {self.dtype} and other: {other.dtype}.  Please validate and cast"
                " as appropriate."
            )

    def add(self, other):
        """
        Define elementwise addition of `DiagMatrix` instances with
        scalars, `BlkDiagMatrix`, and `DiagMatrix`.

        :param other: scalar, `BlkDiagMatrix` or `DiagMatrix.
        :return:  `DiagMatrix` instance with elementwise sum equal
            to self + other. In the case of `BlkDiagMatrix` a `BlkDiagMatrix` is returned.
        """

        if is_scalar_type(other):
            res = DiagMatrix(self._data + other)
        elif isinstance(other, DiagMatrix):
            self.__check_compatible(other)
            res = DiagMatrix(self._data + other._data)
        elif isinstance(other, BlkDiagMatrix):
            # res is blk_diag_matrix
            res = self.as_blk_diag(other.partition) + other
        else:
            raise NotImplementedError(
                f"`add` operation not implemented for {type(other)}"
            )

        return res

    def __add__(self, other):
        """
        Operator overloading for addition.
        """

        return self.add(other)

    def __radd__(self, other):
        """
        Convenience function for elementwise scalar addition.
        """

        return self.add(other)

    def sub(self, other):
        """
        Define elementwise subtraction of `DiagMatrix` instances by
        scalars, `BlkDiagMatrix`, and `DiagMatrix`.

        :param other: scalar, `BlkDiagMatrix` or `DiagMatrix.
        :return:  `DiagMatrix` instance with elementwise sub equal
            to self - other. In the case of `BlkDiagMatrix` a `BlkDiagMatrix` is returned.
        """

        if is_scalar_type(other):
            res = DiagMatrix(self._data - other)
        elif isinstance(other, DiagMatrix):
            self.__check_compatible(other)
            res = DiagMatrix(self._data - other._data)
        elif isinstance(other, BlkDiagMatrix):
            # res is blk_diag_matrix
            res = self.as_blk_diag(other.partition) - other
        else:
            raise NotImplementedError(
                f"`add` operation not implemented for {type(other)}"
            )

        return res

    def __sub__(self, other):
        """
        Operator overloading for subtraction.
        """

        return self.sub(other)

    def __rsub__(self, other):
        """
        Convenience function for elementwise scalar subtraction.
        """

        # Note, the case of DiagMatrix_L - DiagMatrix_R would be
        #   evaluated as L.sub(R), so this is only for other
        #   Object - DiagMatrix situations, namely scalars.

        return -(self - other)

    def matmul(self, other):
        """
        Compute the matrix multiplication of two `DiagMatrix` instances.

        :param other: The rhs `DiagMatrix`, `BlkDiagMatrix` or 2d dense Numpy array.
        :return: Returns `self` @ `other`, as type of `other`
        """
        if isinstance(other, DiagMatrix):
            res = self.mul(other)

        elif isinstance(other, BlkDiagMatrix):
            if self.stack_shape != ():
                raise RuntimeError(
                    f"Mixed `matmul` only implemented for singletons at this time, received {self.stack_shape}."
                )

            res = BlkDiagMatrix.zeros_like(other)

            ind = 0
            for b, blk in enumerate(other):
                i = len(blk)

                res[b] = np.diag(self._data[ind : ind + i]) @ blk
                ind += i

        elif isinstance(other, np.ndarray):
            # For now, lets just interop with 2D `other` arrays,
            # assumed to represent matrices.
            if other.ndim != 2:
                raise ValueError(
                    f"DiagMatrix.matmul of ndarray only supports 2D, received {other.shape}."
                )

            # Then the mathematical `DA` for diag D and matrix A,
            # is expressed in code as `D@A`.
            #
            # The operation `DA` is known as Left Scaling,
            # and means to scale the row A_i by d_i in D.
            # This can be accomplished by broadcasting an elemental multiply.

            res = self._data[..., np.newaxis] * other
        else:
            raise RuntimeError(f"__matmul__ not implemented for {type(other)}.")

        return res

    def __matmul__(self, other):
        """
        Operator overload for matrix multiply of `DiagMatrix` instances.
        """

        return self.matmul(other)

    def __rmatmul__(self, lhs):
        """
        Compute the right matrix multiplication with a `DiagMatrix` instance,
        and a numpy array, `lhs` @ `self`.

        :param other: The lhs Numpy instance.
        :return: Returns numpy array representing `other @ self`.
        """

        # Note, we should only hit this method when mixing DiagMatrix with numpy.
        #   This is because if both a and b are DiagMatrix,
        #   then a@b would be handled first by a.__matmul__(b), never reaching here.
        #   If a is BlkDiagMatrix, then it will handle the conversion of DiagMatrix b.

        if not isinstance(lhs, np.ndarray):
            raise RuntimeError(f"__rmatmul__ not implemented for {type(lhs)}.")

        # For now, lets just interop with 2D `other` arrays,
        # assumed to represent matrices.
        if lhs.ndim != 2:
            raise ValueError(
                f"DiagMatrix.matmul of ndarray only supports 2D, received {lhs.shape}."
            )

        # Then the mathematical `AD` for diag D and matrix A,
        # is expressed in code as `A@D`.
        #
        # The operation `AD` is known as Right Scaling,
        # and means to scale the column A_j by d_j in D.
        # This can be accomplished by broadcasting an elemental multiply.
        # In this case A is the variable `lhs`

        return lhs * self._data[np.newaxis]

    def mul(self, other):
        """
        Compute the elementwise multiplication of a `DiagMatrix` instance and a
        scalar or `DiagMatrix`.

        :param other: The rhs in the multiplication..
        :return: A `self` * `other` as type of `other`.
        """
        if isinstance(other, DiagMatrix):
            self.__check_compatible(other)
            res = DiagMatrix(self._data * other._data)
        elif is_scalar_type(other):  # scalar
            res = DiagMatrix(self._data * other)
        else:
            raise NotImplementedError(f"mul not implemented for {type(other)}.")

        return res

    def __mul__(self, val):
        """
        Operator overload for `DiagMatrix` scalar multiply.
        """

        return self.mul(val)

    def __rmul__(self, other):
        """
        Convenience function, elementwise rmul commutes to mul.
        """
        # `mul` will attempt all compatible multiplications.
        return self.mul(other)

    def neg(self):
        """
        Compute the unary negation of `DiagMatrix` instance.

        :return: A `DiagMatrix` like self.
        """
        return DiagMatrix(-self._data)

    def __neg__(self):
        """
        Operator overload for unary negation of `DiagMatrix` instance.
        """

        return self.neg()

    def abs(self):
        """
        Compute the elementwise absolute value of `DiagMatrix` instance.

        :return: A `DiagMatrix` like self.
        """

        return DiagMatrix(np.abs(self._data))

    def __abs__(self):
        """
        Operator overload for absolute value of `DiagMatrix` instance.
        """

        return self.abs()

    def pow(self, val):
        """
        Compute the elementwise power of `DiagMatrix` instance.

        :param val: Value to exponentiate by.
        :return: A `DiagMatrix` like self.
        """

        return DiagMatrix(self._data**val)

    def __pow__(self, val):
        """
        Operator overload pow of `DiagMatrix` instance.
        """

        return self.pow(val)

    @property
    def norm(self):
        """
        Compute the 2-norm of a `DiagMatrix` instance.

        :return: The 2-norm of the `DiagMatrix` instance.
        """
        # Elements of a diag matrix are its singular values,
        #   and the norm is equal to the largest singular value.
        return np.abs(self._data).max(axis=-1)

    # Transpose methods are provided for reasons of interoperability
    # in code that also uses with BlkDiagMatrix.
    def transpose(self):
        """
        Get the transpose matrix of a `DiagMatrix` instance.

        :return: The corresponding transpose form as a `DiagMatrix`.
        """
        return self.copy()

    @property
    def T(self):
        """
        Syntactic sugar for self.transpose().
        """

        return self.transpose()

    def dense(self):
        """
        Convert `DiagMatrix` instance into full matrix.

        :return: The `DiagMatrix` instance including the zero elements of
        non-diagonal elements.
        """

        if self._data.ndim == 1:
            return np.diag(self._data)

        original_stack_shape = self.stack_shape
        diags = self.stack_reshape(-1)  # Flatten stack

        dense = []
        for d in diags.asnumpy():
            dense.append(np.diag(d))
        # Convert to np.array
        dense = np.concatenate(dense)
        # Reshape to original stack shape.
        return dense.reshape(*original_stack_shape, self.count, self.count)

    def apply(self, X):
        """
        Define the apply option of a diagonal matrix with a matrix of
        coefficient vectors.

        :param X: Coefficient matrix, each column is a coefficient vector.
        :return: A matrix with new coefficient vectors.
        """

        return self * DiagMatrix(X)

    def rapply(self, X):
        """
        Right apply.  Given a matrix of coefficient vectors,
        applies the diagonal matrix on the right hand side.
        Example, X @ self.

        This is the right hand side equivalent to `apply`,
        which due to being diagonal, is the same as `apply`.
        This method exists purely for interoperability with code
        originally targeting `BlkDiagMatrix`.

        :param X: Coefficient matrix, each column is a coefficient vector.

        :return: A matrix with new coefficient vectors.
        """

        return self.apply(X)

    # `eigval` method is provided for reasons of interoperability
    # in code that also uses with `BlkDiagMatrix`.
    def eigvals(self):
        """
        Compute the eigenvalues of a `DiagMatrix`.

        :return: Array of eigvals, with length equal to the fully expanded matrix diagonal.
        """
        return self.asnumpy()

    @staticmethod
    def empty(shape, dtype=np.float32):
        """
        Instantiate an empty `DiagMatrix` with `shape`.
        When shape is an integer, this corresponds to
        the diag(A) where A is (n,n).
        Note, like Numpy, empty values are uninitialized.

        :param shape: Shape of matrix.
            When integer, corresponds to len(diag(A)).  Otherwise,
            when a tuple (..., n), the last dimension `n` defines the
            length of diagonal, while prior dimensions define any
            stack axes.
        :param dtype: Datatype, defaults to np.float32.
        :return: `DiagMatrix` instance.
        """

        return DiagMatrix(np.empty(shape, dtype=dtype))

    @staticmethod
    def zeros(shape, dtype=np.float32):
        """
        Instantiate a zero intialized `DiagMatrix`.
        When `shape` is an integer, this corresponds to
        the diag(A) where A is (n,n).

        :param shape: Shape of matrix.
            When integer, corresponds to len(diag(A)).  Otherwise,
            when a tuple (..., n), the last dimension `n` defines the
            length of diagonal, while prior dimensions define any
            stack axes.
        :param dtype: Datatype, defaults to np.float32.
        :return: `DiagMatrix` instance.
        """

        return DiagMatrix(np.zeros(shape, dtype=dtype))

    @staticmethod
    def ones(shape, dtype=np.float32):
        """
        Instantiate ones intialized `DiagMatrix`.
        When `shape` is an integer, this corresponds to
        the diag(A) where A is (n,n).

        :param shape: Shape of matrix.
            When integer, corresponds to len(diag(A)).  Otherwise,
            when a tuple (..., n), the last dimension `n` defines the
            length of diagonal, while prior dimensions define any
            stack axes.
        :param dtype: Datatype, defaults to np.float32.
        :return: `DiagMatrix` instance.
        """

        return DiagMatrix(np.ones(shape, dtype=dtype))

    @staticmethod
    def eye(shape, dtype=np.float32):
        """
        Build a `DiagMatrix` eye (identity) matrix.
        This is simply an alias for `ones`.

        :param shape: Shape of matrix.
            When integer, corresponds to len(diag(A)).  Otherwise,
            when a tuple (..., n), the last dimension `n` defines the
            length of diagonal, while prior dimensions define any
            stack axes.
        :param dtype: Datatype, defaults to np.float32.
        :return: `DiagMatrix` instance.
        """

        return DiagMatrix.ones(shape, dtype=dtype)

    def as_blk_diag(self, partition):
        """
        Express `DiagMatrix` as a `BlkDiagMatrix` using `partition`.

        :param partition: BlkDiagMatrix partition.
        :return: `BlkDiagMatrix`
        """
        if self.stack_shape != ():
            raise RuntimeError(
                f"as_blk_diag only implemented for singletons at this time, received {self.stack_shape}."
            )

        B = BlkDiagMatrix(partition, dtype=self.dtype)
        ind = 0
        for b, p in enumerate(partition):
            if p[0] != p[1]:
                raise RuntimeError(
                    f"Block {b} of partition is not square, {p}."
                    "Currently DiagMatrix.as_blk_diag only supports square blocks."
                )

            j = p[0]
            B[b] = np.diag(self._data[ind : ind + j])
            ind += j

        return B

    def solve(self, b):
        """
        For this `DiagMatrix` `a` and vector `b`.
        solve a x = b for `x`.

        :param b: Right hand side, Numpy array.
        :return: `DiagMatrix`, solution `x`.
        """

        if self.stack_shape != ():
            raise RuntimeError(
                f"`solve` only implemented for singletons at this time, received {self.stack_shape}."
            )

        return DiagMatrix(b / self._data)



"""
This code is an ASPIRE port of Ruiyi Yang's `Bayesian Optimal Transport Alignment`.

`https://github.com/RuiyiYang/BOTalign`

`https://arxiv.org/pdf/2305.12310.pdf`
"""

import numpy as np
import pymanopt
from numpy.linalg import norm
from scipy.optimize import minimize

from aspire.operators import wemd_embed
from aspire.utils.rotation import Rotation

# Store parameters specific to each loss_type.
# `lengthscale` is used to scale the modeled covariance
# functions, corresponding to equations 7 through 9 in the paper.
# `wavelet` and `level` are passed directly to `wemd_embed`.
loss_types = {
    "wemd": dict(lengthscale=0.75, wavelet="sym3", level=6),
    "euclidean": dict(lengthscale=1),
}


def align_BO(
    vol_ref,
    vol_given,
    loss_type="wemd",
    loss_params=None,
    downsampled_size=32,
    refinement_downsampled_size=32,
    max_iters=200,
    refine=True,
    tau=1e-3,
    surrogate_max_iter=500,
    surrogate_min_grad=0.1,
    surrogate_min_step=0.1,
    verbosity=0,
    dtype=None,
):
    """
    This function returns a rotation matrix R that best aligns vol_ref with the rotated version of vol_given.

    :param vol_ref: The reference `Volume`
    :param vol_given: The `Volume` to be aligned
    :param loss_type: 'wemd' or 'euclidean'. Default 'wemd'.
        If the heterogeneity between vol_ref and vol_given is high, 'euclidean' is recommended.
    :param loss_params: Optional dictionary for overriding parameters in `aspire.utils.bot_align.loss_types`.
        Defaults to empty dictionary.
    :param downsampled_size: Downsampling (pixels). Integer, defaults to 32.
        If alignment fails try larger values.
    :param refinement_downsampled_size: Downsampling (pixels) used with refinement. Integer, defaults to 32.
    :param max_iters: Maximum iterations. Integer, defaults 200.
        If alignment fails try larger values.
    :param refine: Whether to perform refinement. Boolean, defaults True.
    :param tau: Regularization parameter for surrogate problems. Numeric, defaults 1e-3.
    :param surrogate_max_iter: Stopping criterion for surrogate problems--maximum iterations. Interger, defaults 500.
    :param surrogate_min_grad: Stopping criterion for surrogate problems--minimum gradient norm. Numeric, defaults 0.1.
    :param surrogate_min_step: Stopping criterion for surrogate problems--minimum step size. Numeric, defaults 0.1.
    :param verbosity: Surrogate problem optimization detail level. integer, defaults 0 (silent). 2 is most verbose.
    :param dtype: Numeric dtype to perform computations with.
        Default `None` infers dtype from `vol_ref`.
    :return: Rotation matrix R_init (without refinement) or (R_init, R_est) (with refinement).
    """
    # Infer dtype
    dtype = np.dtype(dtype or vol_ref.dtype)

    # Raise error on volume dtype mismatch.
    if vol_ref.dtype != dtype:
        raise ValueError(f"`vol_ref` dtype {vol_ref.dtype} != {dtype}.")
    if vol_given.dtype != dtype:
        raise ValueError(f"`vol_ref` dtype {vol_given.dtype} != {dtype}.")

    loss_type = loss_type.lower()
    if loss_type not in loss_types.keys():
        raise ValueError(f"BOTalign `loss_type` must be one of {loss_types.keys()}")

    # Update defaults param items with any user specified items.
    if loss_params is None:
        loss_params = dict()
    loss_params = {**loss_types[loss_type], **loss_params}

    # Downsample Volumes
    vol_ref_ds = vol_ref.downsample(downsampled_size)
    vol_given_ds = vol_given.downsample(downsampled_size)

    if loss_type == "wemd":
        embed_0 = wemd_embed(
            vol_ref_ds.asnumpy()[0], loss_params["wavelet"], loss_params["level"]
        )

    def loss_fun(R):
        """
        Alignment loss function.
        """
        v = vol_given_ds.rotate(Rotation(R)).asnumpy()[0]
        if loss_type == "euclidean":
            return norm(v - vol_ref_ds)
        elif loss_type == "wemd":
            embed_v = wemd_embed(v, loss_params["wavelet"], loss_params["level"])
            return norm(embed_v - embed_0, ord=1)

    def cov_fun(x1, x2):
        """
        The squared exponential covariance function over SO(3).
        """
        d = norm(x1 - x2) / loss_params["lengthscale"]
        return np.exp(-(d**2) / 2, dtype=dtype)

    def cov_fun_grad(x1, x2):
        """
        Surrogate gradient of the squared exponential covariance function,
        with respect to `x1`.
        """
        # Corresponds to equation 11 in the paper.
        return cov_fun(x1, x2) * (x2 - x1) / loss_params["lengthscale"] ** 2

    R = np.zeros((max_iters, 3, 3), dtype=dtype)
    R[0] = np.eye(3, dtype=dtype)
    manifold = pymanopt.manifolds.SpecialOrthogonalGroup(3)

    cov = np.zeros((max_iters, max_iters), dtype=dtype)
    cov[0, 0] = cov_fun(R[0], R[0])

    loss = np.zeros(max_iters, dtype=dtype)
    loss[0] = loss_fun(R[0])

    for t in range(1, max_iters):
        # See discussion 3.2 and equation 10 in the paper.
        q = np.linalg.solve(cov[:t, :t] + tau * np.eye(t, dtype=dtype), loss[:t])

        @pymanopt.function.numpy(manifold)
        def cost(new, t=t, q=q):
            """
            Loss function for surrogate problems.
            """
            kx = np.array(
                [cov_fun(new.astype(dtype, copy=False), R[j]) for j in range(t)]
            )
            mu = kx @ q
            return mu

        @pymanopt.function.numpy(manifold)
        def euclidean_grad(new, t=t, q=q):
            """
            Gradient of the loss function for surrogate problems.
            """
            # Corresponds to equation 11 in the paper.
            kx_grad = np.array(
                [cov_fun_grad(new.astype(dtype, copy=False), R[j]) for j in range(t)]
            )

            kx_grad = kx_grad.reshape((t, 9))
            grad = q @ kx_grad
            grad = grad.reshape((3, 3))

            return grad

        problem = pymanopt.Problem(manifold, cost, euclidean_gradient=euclidean_grad)
        optimizer = pymanopt.optimizers.SteepestDescent(
            max_iterations=surrogate_max_iter,
            min_gradient_norm=surrogate_min_grad,
            min_step_size=surrogate_min_step,
            verbosity=verbosity,
        )
        result = optimizer.run(problem)
        R_new = result.point.astype(dtype, copy=False)

        loss[t] = loss_fun(R_new)
        R[t] = R_new
        k = np.array([cov_fun(R_new, R[i]) for i in range(t + 1)])
        cov[t, 0:t] = k[0:t]
        cov[0 : t + 1, t] = k

    idx = loss.argmin(0)
    R_init = R[idx]

    if downsampled_size > refinement_downsampled_size:
        vol_ref_ds = vol_ref_ds.downsample(refinement_downsampled_size)
        vol_given_ds = vol_given_ds.downsample(refinement_downsampled_size)

    if refine:
        sign = np.sign(np.linalg.det(R_init))

        def loss_u(u):
            """
            Convert the alignment loss function over SO(3) to be over R^3.
            """
            u = u.astype(dtype, copy=False)
            R = sign * Rotation.from_rotvec(u).matrices[0]
            v_rot = vol_given_ds.rotate(Rotation(R)).asnumpy()[0]
            return norm(vol_ref_ds - v_rot)

        x0 = Rotation(sign * R_init).as_rotvec()[0]
        result = minimize(loss_u, x0, method="nelder-mead", options={"disp": False})
        u_est = result.x
        R_est = sign * Rotation.from_rotvec(u_est).matrices[0]

        return R_init, R_est
    else:
        return R_init



# dataset registry for ASPIRE example data.
registry = {
    "emdb_2660.map": "49aecfd4efce09afc937d1786bbed6f18c2a353c73a4e16a643a304342d0660e",
    "emdb_8012.map": "85a1c9ab958b1dd051d011515212d58decaf2537351b9c016acd3e5852e30d63",
    "emdb_2984.map": "0194f44cb28f2a8daa5d477d25852de9cc81ed093487de181ee4b30b0d77ef90",
    "emdb_8511.map": "1f03ec4a0cadb407b6b972c803ffe1e97ff5087d4c2ce9fec2c404747a7fb3fe",
    "emdb_3645.map": "a574dba9657d44665b87f368f7379e97b1a33fe6ac2540478a3707f5ea840f12",
    "emdb_4905.map": "fe9ce303b43b11ccf253c8786b339cea3273ef70ff49dd6099155d576181f3c2",
    "emdb_10835.map": "dec12cdda4c36253a15f6f8105416020855bd51eb860ac5eb04b37b000ac9156",
    "emdb_5778.map": "877cbe37b86561c3dfb255aa2308fefcdd8f51f91928b17c2ef5c8dd3afaaef7",
    "emdb_6287.map": "81463aa6d024c80efcd19aa9b5ac58f3b3464af56e1ef0f104bd25071acc9204",
    "emdb_2824.map": "7682e1ef6e5bc9f2de9edcf824a03e454ef9cb1ca33bc12920633559f7f826e4",
    "emdb_14621.map": "b45774245c2bd5e1a44e801b8fb1705a44d5850631838d060294be42e34a6900",
    "emdb_2484.map": "6a324e23352bea101c191d5e854026162a5a9b0b8fc73ac5a085cc22038e1999",
    "emdb_6458.map": "645208af6d36bbd3d172c549e58d387b81142fd320e064bc66105be0eae540d1",
}

registry_urls = {
    "emdb_2660.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-2660/map/emd_2660.map.gz",
    "emdb_8012.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-8012/map/emd_8012.map.gz",
    "emdb_2984.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-2984/map/emd_2984.map.gz",
    "emdb_8511.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-8511/map/emd_8511.map.gz",
    "emdb_3645.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-3645/map/emd_3645.map.gz",
    "emdb_4905.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-4905/map/emd_4905.map.gz",
    "emdb_10835.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-10835/map/emd_10835.map.gz",
    "emdb_5778.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-5778/map/emd_5778.map.gz",
    "emdb_6287.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-6287/map/emd_6287.map.gz",
    "emdb_2824.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-2824/map/emd_2824.map.gz",
    "emdb_14621.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-14621/map/emd_14621.map.gz",
    "emdb_2484.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-2484/map/emd_2484.map.gz",
    "emdb_6458.map": "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-6458/map/emd_6458.map.gz",
}

file_to_method_map = {
    "emdb_2660.map": "emdb_2660",
    "emdb_8012.map": "emdb_8012",
    "emdb_2984.map": "emdb_2984",
    "emdb_8511.map": "emdb_8511",
    "emdb_3645.map": "emdb_3645",
    "emdb_4905.map": "emdb_4905",
    "emdb_10835.map": "emdb_10835",
    "emdb_5778.map": "emdb_5778",
    "emdb_6287.map": "emdb_6287",
    "emdb_2824.map": "emdb_2824",
    "emdb_14621.map": "emdb_14621",
    "emdb_2484.map": "emdb_2484",
    "emdb_6458.map": "emdb_6458",
}



import logging

from aspire.reconstruction.kernel import FourierKernel

logger = logging.getLogger(__name__)


class Estimator:
    def __init__(self, src, basis, batch_size=512, preconditioner="circulant"):
        """
        An object representing a 2*L-by-2*L-by-2*L array containing the non-centered Fourier transform of the mean
        least-squares estimator kernel.
        Convolving a volume with this kernel is equal to projecting and backproject-ing that volume in each of the
        projection directions (with the appropriate amplitude multipliers and CTFs) and averaging over the whole
        dataset.
        Note that this is a non-centered Fourier transform, so the zero frequency is found at index 1.
        """

        self.src = src
        self.basis = basis
        self.dtype = self.src.dtype
        self.batch_size = batch_size
        self.preconditioner = preconditioner

        if not self.dtype == self.basis.dtype:
            logger.warning(
                f"Inconsistent types in {self.dtype} Estimator."
                f" basis: {self.basis.dtype}"
            )

        if src.L != basis.nres:
            raise ValueError(
                "Currently require 2D source and 3D volume resolution to be the same."
                f" Given src.L={src.L} != {basis.nres}"
            )

    def __getattr__(self, name):
        """Lazy attributes instantiated on first-access"""

        if name == "kernel":
            logger.info("Computing kernel")
            kernel = self.kernel = self.compute_kernel()
            return kernel

        elif name == "precond_kernel":
            if self.preconditioner == "circulant":
                logger.info("Computing Preconditioner kernel")
                precond_kernel = self.precond_kernel = FourierKernel(
                    1.0 / self.kernel.circularize()
                )
            else:
                precond_kernel = self.precond_kernel = None
            return precond_kernel

        else:
            raise AttributeError(name)

    def compute_kernel(self):
        raise NotImplementedError("Subclasses must implement the compute_kernel method")

    def estimate(self, b_coeff=None, tol=1e-5, regularizer=0):
        """Return an estimate as a Volume instance."""
        if b_coeff is None:
            b_coeff = self.src_backward()
        est_coeff = self.conj_grad(b_coeff, tol=tol, regularizer=regularizer)
        est = self.basis.evaluate(est_coeff).T

        return est

    def apply_kernel(self, vol_coeff, kernel=None):
        """
        Applies the kernel represented by convolution
        :param vol_coeff: The volume to be convolved, stored in the basis coefficients.
        :param kernel: a Kernel object. If None, the kernel for this Estimator is used.
        :return: The result of evaluating `vol_coeff` in the given basis, convolving with the kernel given by
            kernel, and backprojecting into the basis.
        """
        if kernel is None:
            kernel = self.kernel
        vol = self.basis.evaluate(vol_coeff)  # returns a Volume
        vol = kernel.convolve_volume(vol)  # returns a Volume
        vol_coef = self.basis.evaluate_t(vol)
        return vol_coef



import click

from aspire import config
from aspire.utils.logging import LOGGING_LEVEL_NAMES

# universal option for aspire commands
log_level_option = click.option(
    "--loglevel",
    type=click.Choice(LOGGING_LEVEL_NAMES, case_sensitive=False),
    default=config["logging"]["console_level"].as_choice(LOGGING_LEVEL_NAMES),
    help="Logging verbosity level of console output.",
)



import logging

import numpy as np

from aspire.basis import Basis, FBBasisMixin
from aspire.basis.basis_utils import real_sph_harmonic, sph_bessel, unique_coords_nd
from aspire.utils import roll_dim, unroll_dim
from aspire.utils.matlab_compat import m_flatten, m_reshape

logger = logging.getLogger(__name__)


class FBBasis3D(Basis, FBBasisMixin):
    """
    Define a derived class for direct spherical Harmonics Bessel basis expanding 3D volumes

    # TODO: Methods that return dictionaries should return useful objects instead

    """

    def __init__(self, size, ell_max=None, dtype=np.float32):
        """
        Initialize an object for the 3D Fourier-Bessel basis class

        :param size: The size of the vectors for which to define the basis.
            May be a 3-tuple or an integer, in which case a cubic basis is assumed.
            Currently only cubic images are supported.
        :param ell_max: The maximum order ell of the basis elements. If no input
            (= None), it will be set to np.Inf and the basis includes all
            ell such that the resulting basis vectors are concentrated
            below the Nyquist frequency (default Inf).
        """
        if isinstance(size, int):
            size = (size, size, size)
        ndim = len(size)
        assert ndim == 3, "Only three-dimensional basis functions are supported."
        assert len(set(size)) == 1, "Only cubic domains are supported."

        super().__init__(size, ell_max, dtype=dtype)

    def _build(self):
        """
        Build the internal data structure for 3D Fourier-Bessel basis
        """

        logger.info(
            "Expanding 3D map in a spatial-domain Fourier–Bessel"
            " basis using the direct method."
        )

        # get upper bound of zeros, ells, and ks  of Bessel functions
        self._calc_k_max()

        # calculate total number of basis functions
        self.count = sum(self.k_max * (2 * np.arange(0, self.ell_max + 1) + 1))

        # obtain a 3D grid to represent basis functions
        self.basis_coords = unique_coords_nd(self.nres, self.ndim, dtype=self.dtype)

        # generate 1D indices for basis functions
        self._indices = self.indices()

        # precompute the basis functions in 3D grids
        self._precomp = self._precomp()

        # get normalized factors
        self._norms = self.norms()

    def indices(self):
        """
        Create the indices for each basis function
        """
        indices_ells = np.zeros(self.count, dtype=int)
        indices_ms = np.zeros(self.count, dtype=int)
        indices_ks = np.zeros(self.count, dtype=int)

        ind = 0
        for ell in range(self.ell_max + 1):
            ks = range(0, self.k_max[ell])
            for m in range(-ell, ell + 1):
                rng = range(ind, ind + len(ks))
                indices_ells[rng] = ell
                indices_ms[rng] = m
                indices_ks[rng] = ks

                ind += len(ks)

        return {"ells": indices_ells, "ms": indices_ms, "ks": indices_ks}

    def _precomp(self):
        """
        Precompute the basis functions at defined sample points
        """
        r_unique = self.basis_coords["r_unique"]
        ang_unique = self.basis_coords["ang_unique"]

        ind_radial = 0
        ind_ang = 0

        radial = np.zeros(shape=(len(r_unique), np.sum(self.k_max)), dtype=self.dtype)
        ang = np.zeros(
            shape=(ang_unique.shape[-1], (self.ell_max + 1) ** 2), dtype=self.dtype
        )

        for ell in range(0, self.ell_max + 1):
            for k in range(1, self.k_max[ell] + 1):
                radial[:, ind_radial] = sph_bessel(ell, self.r0[ell][k - 1] * r_unique)
                ind_radial += 1

            for m in range(-ell, ell + 1):
                ang[:, ind_ang] = real_sph_harmonic(
                    ell, m, ang_unique[0, :], ang_unique[1, :]
                )
                ind_ang += 1

        return {"radial": radial, "ang": ang}

    def norms(self):
        """
        Calculate the normalized factors of basis functions
        """
        norms = np.zeros(np.sum(self.k_max), dtype=self.dtype)
        norm_fn = self.basis_norm_3d

        i = 0
        for ell in range(0, self.ell_max + 1):
            for k in range(1, self.k_max[ell] + 1):
                norms[i] = norm_fn(ell, k)
                i += 1

        return norms

    def basis_norm_3d(self, ell, k):
        """
        Calculate the normalized factor of a specified basis function.
        """
        return (
            np.abs(sph_bessel(ell + 1, self.r0[ell][k - 1]))
            / np.sqrt(2)
            * np.sqrt((self.nres / 2) ** 3)
        )

    def _evaluate(self, v):
        """
        Evaluate coefficients in standard 3D coordinate basis from those in FB basis

        :param v: A coefficient vector (or an array of coefficient vectors) to
            be evaluated. The first dimension must equal `self.count`.
        :return: The evaluation of the coefficient vector(s) `v` for this basis.
            This is an array whose first dimensions equal `self.z` and the
            remaining dimensions correspond to dimensions two and higher of `v`.
        """

        v = v.T
        v, sz_roll = unroll_dim(v, 2)

        r_idx = self.basis_coords["r_idx"]
        ang_idx = self.basis_coords["ang_idx"]
        mask = m_flatten(self.basis_coords["mask"])

        ind = 0
        ind_radial = 0
        ind_ang = 0

        x = np.zeros(
            shape=tuple([np.prod(self.sz)] + list(v.shape[1:])), dtype=self.dtype
        )
        for ell in range(0, self.ell_max + 1):
            k_max = self.k_max[ell]
            idx_radial = ind_radial + np.arange(0, k_max)
            nrms = self._norms[idx_radial]
            radial = self._precomp["radial"][:, idx_radial]
            radial = radial / nrms

            for _ in range(-ell, ell + 1):
                ang = self._precomp["ang"][:, ind_ang]
                ang_radial = np.expand_dims(ang[ang_idx], axis=1) * radial[r_idx]
                idx = ind + np.arange(0, len(idx_radial))
                x[mask] += ang_radial @ v[idx]
                ind += len(idx)
                ind_ang += 1

            ind_radial += len(idx_radial)

        x = m_reshape(x, self.sz + x.shape[1:])
        x = roll_dim(x, sz_roll)

        return x.T

    def _evaluate_t(self, v):
        """
        Evaluate coefficient in FB basis from those in standard 3D coordinate basis

        :param v: The coefficient array to be evaluated. The first dimensions
            must equal `self.sz`.
        :return: The evaluation of the coefficient array `v` in the dual
            basis of `basis`. This is an array of vectors whose first dimension
            equals `self.count` and whose remaining dimensions correspond
            to higher dimensions of `v`.
        """
        v = v.T
        x, sz_roll = unroll_dim(v, self.ndim + 1)
        x = m_reshape(
            x, new_shape=tuple([np.prod(self.sz)] + list(x.shape[self.ndim :]))
        )

        r_idx = self.basis_coords["r_idx"]
        ang_idx = self.basis_coords["ang_idx"]
        mask = m_flatten(self.basis_coords["mask"])

        ind = 0
        ind_radial = 0
        ind_ang = 0

        v = np.zeros(shape=tuple([self.count] + list(x.shape[1:])), dtype=self.dtype)
        for ell in range(0, self.ell_max + 1):
            k_max = self.k_max[ell]
            idx_radial = ind_radial + np.arange(0, k_max)
            nrms = self._norms[idx_radial]
            radial = self._precomp["radial"][:, idx_radial]
            radial = radial / nrms

            for _ in range(-ell, ell + 1):
                ang = self._precomp["ang"][:, ind_ang]
                ang_radial = np.expand_dims(ang[ang_idx], axis=1) * radial[r_idx]
                idx = ind + np.arange(0, len(idx_radial))
                v[idx] = np.real(ang_radial.T @ x[mask])
                ind += len(idx)
                ind_ang += 1

            ind_radial += len(idx_radial)

        v = roll_dim(v, sz_roll)
        return v.T



import logging

import numpy as np

from aspire.basis import FFBBasis2D
from aspire.classification import (
    Averager2D,
    BFRAverager2D,
    BFSRAverager2D,
    Class2D,
    ClassSelector,
    NeighborVarianceWithRepulsionClassSelector,
    RIRClass2D,
    TopClassSelector,
)
from aspire.image import Image
from aspire.source import ImageSource

logger = logging.getLogger(__name__)


class ClassAvgSource(ImageSource):
    """
    Source for denoised 2D images using class average methods.
    """

    def __init__(
        self,
        src,
        classifier,
        class_selector,
        averager,
    ):
        """
        Constructor of an object for denoising 2D images using class averaging methods.

        :param src: Source used for image classification.
        :param classifier: Class2D subclass used for image classification.
            Example, RIRClass2D.
        :param class_selector: A ClassSelector subclass.
        :param averager: An Averager2D subclass.
        """
        self.src = src
        if not isinstance(self.src, ImageSource):
            raise ValueError(
                f"`src` should be subclass of `ImageSource`, found {self.src}."
            )

        self.classifier = classifier
        if not isinstance(self.classifier, Class2D):
            raise ValueError(
                f"`classifier` should be subclass of `Class2D`, found {self.classifier}."
            )

        self.class_selector = class_selector
        if not isinstance(self.class_selector, ClassSelector):
            raise ValueError(
                f"`class_selector` should be instance of `ClassSelector`, found {class_selector}."
            )

        self.averager = averager
        if not isinstance(self.averager, Averager2D):
            raise ValueError(
                f"`averager` should be instance of `Averager2D`, found {self.averager}."
            )

        # Flag for lazy eval, we'll classify once, on first touch.
        self._classified = False
        self._selected = False

        # Note n will potentially be updated after class selection.
        # Manage delayed setting `n` once.
        self._n_set = False

        super().__init__(
            L=self.averager.src.L,
            n=self.averager.src.n,
            dtype=self.averager.src.dtype,
        )

        # Any further operations should not mutate this instance.
        # Note, updating `n` is a special case for this source at this time.
        self._mutable = False

    @ImageSource.n.setter
    def n(self, n):
        """
        Sets max image index `n` in `src` and associated
        `ImageAccessor`.

        :param n: Number of images.
        """

        # Resets the protection of self._n exactly once after __init__.
        if (self._n is not None) and (not self._n_set):
            self._n = None
            self._n_set = True

        super()._set_n(n)

    def _classify(self):
        """
        Perform the image classification (if not already done).
        """

        # Short circuit
        if self._classified:
            logger.debug(f"{self.__class__.__name__} already classified, skipping")
            return

        (
            self.class_indices,
            self.class_refl,
            self.class_distances,
        ) = self.classifier.classify()
        self._classified = True

    @property
    def selection_indices(self):
        self._class_select()
        return super().selection_indices

    @selection_indices.setter
    def selection_indices(self, value):
        self.set_metadata(["_selection_indices"], value)

    @property
    def class_indices(self):
        """
        Returns table of class image indices as `(src.n, n_nbors)`
        Numpy array.

        Each row reprsents a class, with the columns ordered by
        smallest `class_distances` from the reference image (zeroth
        columm).

        Note `n_nbors` is managed by `self.classifier` and used here
        for documentation.

        :return: Numpy array, integers.
        """
        self._classify()
        return super().class_indices

    @class_indices.setter
    def class_indices(self, table):
        self.set_metadata(
            ["_class_indices"], [",".join(map(str, row)) for row in table]
        )

    @property
    def class_refl(self):
        """
        Returns table of class image reflections as `(src.n, n_nbors)`
        Numpy array.

        Follows same layout as `class_indices` but holds booleans that
        are True when the image should be reflected before averaging.

        Note `n_nbors` is managed by `self.classifier` and used here
        for documentation.

        :return: Numpy array, boolean.
        """
        self._classify()
        return super().class_refl

    @class_refl.setter
    def class_refl(self, table):
        # Convert boolean to (O, 1) integers.
        array_int = np.array(table, dtype=int)
        self.set_metadata(
            ["_class_refl"], [",".join(map(str, row)) for row in array_int]
        )

    @property
    def class_distances(self):
        """
        Returns table of class image distances as `(src.n, n_nbors)`
        Numpy array.

        Follows same layout as `class_indices` but holds floats
        representing the distance (returned by classifier) to the
        zeroth image in each class.

        Note `n_nbors` is managed by `self.classifier` and used here
        for documentation.

        :return: Numpy array, self.dtype.
        """
        self._classify()
        return super().class_distances

    @class_distances.setter
    def class_distances(self, table):
        self.set_metadata(
            ["_class_distances"], [",".join(map(str, row)) for row in table]
        )

    def _class_select(self):
        """
        Uses the `class_selector` in conjunction with the classifier results
        to select the classes (and order) used for image averager,
        if not already done.
        """

        # Short circuit
        if self._selected:
            logger.debug(
                f"{self.__class__.__name__} already selected classes, skipping"
            )
            return

        # Evaluate the classification network.
        if not self._classified:
            self._classify()

        # Perform class selection
        _selection_indices = self.class_selector.select(
            self.class_indices,
            self.class_refl,
            self.class_distances,
        )

        # Override the initial self.n
        # Some selectors will (dramatically) reduce the space of classes.
        if len(_selection_indices) != self.n:
            logger.info(
                f"After selection process, updating maximum {len(_selection_indices)} classes from {self.n}."
            )
        # Note, setter should be inherited from base ImageSource.
        self.n = len(_selection_indices)

        self._selected = True
        self.selection_indices = _selection_indices

    def save(self, *args, **kwargs):
        """
        Save metadata to STAR file.

        See `ImageSource.save` for documentation.
        """
        # Evaluate any lazy actions.
        # This should populate relevant metadata.
        self._class_select()
        # Call parent `save` method.
        return super().save(*args, **kwargs)

    def _images(self, indices):
        """
        Output images
        """

        # Lazy evaluate the class selection
        if not self._selected:
            self._class_select()

        # Truncate the request if nessecary,
        # ie, when selection reduces `self.n`.
        indices = np.array(indices, dtype=int)
        selected = indices[indices < self.n]

        if len(indices) != len(selected):
            deselected = indices[indices >= self.n]
            logger.debug(
                f"Dropping requested indices {deselected} following to selection process."
            )
        indices = selected

        # Remap to the selected ordering
        _indices = indices.copy()  # Store original mapping
        indices = self.selection_indices[indices]

        # Check if there is a cache available from class selection component.
        # Note, we can use := for this in the branch directly, when Python>=3.8
        heap_inds = None
        # Check we are using the same averager before attempting to use heap.
        if hasattr(self.class_selector, "heap") and (
            self.averager == self.class_selector.averager
        ):
            # Then check if request matches anything in the heap.
            heap_inds = set(indices).intersection(self.class_selector.heap_ids)

        # Check if this src cached images.
        if self._cached_im is not None:
            logger.debug(f"Loading {len(indices)} images from image cache")
            im = Image(self._cached_im[indices, :, :])

        # Check for heap cached image sets from class_selector.
        elif heap_inds:
            logger.debug(f"Mapping {len(heap_inds)} images from heap cache.")

            # Images in heap_inds can be fetched from class_selector.
            # For others, create an indexing map that preserves
            # original order.  Both of these dicts map requested image
            # id to location in the return array `im`.

            # Note, this stores inds from the remapped `indices`.
            indices_from_heap = {
                ind: i for i, ind in enumerate(indices) if ind in heap_inds
            }
            # Note, this stores _inds from the original `_indices`.
            # This allows the recusive call, which remaps the indices.
            indices_to_compute = {
                _ind: i
                for i, _ind in enumerate(_indices)
                if self.selection_indices[_ind] not in heap_inds
            }

            # Get heap dict once to avoid traversing heap in a loop.
            heap_dict = self.class_selector.heap_idx_map

            # Create an empty array to pack results.
            L = self.averager.src.L
            im = np.empty(
                (len(indices), L, L),
                dtype=self.averager.dtype,
            )

            # Recursively call `_images`.
            # `heap_inds` set should be empty in the recursive call,
            # and compute only remaining images (those not in heap).
            _compute_indices = list(indices_to_compute.keys())
            # Skip when empty (everything requested in heap).
            if len(_compute_indices):
                _imgs = self._images(_compute_indices)

                # Pack images computed from `_images` recursive call.
                _inds = list(indices_to_compute.values())
                im[_inds] = _imgs

            # Pack images from heap.
            for k, i in indices_from_heap.items():
                # map the image index to heap item location
                heap_loc = heap_dict[k]
                im[i] = self.class_selector.heap[heap_loc].image

            # Finally construct an Image.
            im = Image(im)

        else:
            # Perform image averaging for the requested images (classes)
            logger.debug(f"Averaging {len(_indices)} images from source")
            im = self.averager.average(
                self.class_indices[_indices], self.class_refl[_indices]
            )

        # Finally, apply transforms to resulting Images
        return self.generation_pipeline.forward(im, indices)


# The following sub classes attempt to pack sensible defaults
#   into ClassAvgSource so that users don't need to
#   instantiate every component to get started.


class DebugClassAvgSource(ClassAvgSource):
    """
    Source for denoised 2D images using class average methods.

    In this context Debug means defaulting to:
        * Using the defaults for `RIRClass2D`.
        * Using `TopClassSelector` to select all classes maintaining the same order as the input source.
        * Using `BFRAverager2D` with defaults on a single core.
    """

    def __init__(
        self,
        src,
        n_nbor=10,
        num_procs=1,  # Change to "auto" if your machine has many processors
        classifier=None,
        class_selector=None,
        averager=None,
    ):
        """
        Instantiates with default debug paramaters.

        :param src: Source used for image classification.
        :param n_nbor: Number of nearest neighbors. Default 10.
        :param num_procs: Number of processors. Default of 1 runs serially.
            `None` attempts to compute a reasonable value
            based on available cores and memory.
        :param classifier: `Class2D` classifier instance.
            Default `None` creates `RIRClass2D`.
            See code for parameter details.
        :param class_selector: `ClassSelector` instance.
            Default `None` creates `TopClassSelector`.
        :param averager: `Averager2D` instance.
            Default `None` ceates `BFRAverager2D` instance.
            See code for parameter details.

        :return: ClassAvgSource instance.
        """
        dtype = src.dtype

        if classifier is None:
            classifier = RIRClass2D(
                src,
                fspca_components=400,
                bispectrum_components=300,  # Compressed Features after last PCA stage.
                n_nbor=n_nbor,
                large_pca_implementation="legacy",
                nn_implementation="legacy",
                bispectrum_implementation="legacy",
            )

        if averager is None:
            averager = BFRAverager2D(
                FFBBasis2D(src.L, dtype=src.dtype),
                src,
                num_procs=num_procs,
                dtype=dtype,
            )

        if class_selector is None:
            class_selector = TopClassSelector()

        super().__init__(
            src=src,
            classifier=classifier,
            class_selector=class_selector,
            averager=averager,
        )


def DefaultClassAvgSource(
    src,
    n_nbor=50,
    num_procs=None,
    classifier=None,
    class_selector=None,
    averager=None,
    averager_src=None,
    version=None,
):
    """
    Source for denoised 2D images using class average methods.

    Accepts `version`, to dispatch ClassAvgSource with parameters
    below.  Default `version` is latest available.

    :param src: Source used for image classification.
    :param n_nbor: Number of nearest neighbors. Default 50.
    :param num_procs: Number of processors. Use 1 to run serially.
        Default `None` attempts to compute a reasonable value
        based on available cores and memory.
    :param classifier: `Class2D` classifier instance.
        Default `None` creates `RIRClass2D`.
        See code for parameter details.
    :param class_selector: `ClassSelector` instance.
        Default `None` creates `NeighborVarianceWithRepulsionClassSelector`.
    :param averager: `Averager2D` instance.
        Default `None` ceates `BFSRAverager2D` instance.
        See code for parameter details.
    :param averager_src: Optionally explicitly assign source
        to BFSRAverager2D during initialization.
        Raises error when combined with an explicit `averager`
        argument.
    :param version: Optionally selects a versioned DefaultClassAvgSource.
        Defaults to latest available.
    :return: ClassAvgSource instance.
    """

    _versions = {
        None: ClassAvgSourcev110,
        "latest": ClassAvgSourcev110,
        "0.11.0": ClassAvgSourcev110,
    }

    if version not in _versions:
        raise RuntimeError(f"DefaultClassAvgSource version {version} not found.")
    cls = _versions[version]

    return cls(
        src,
        n_nbor=n_nbor,
        num_procs=num_procs,
        classifier=classifier,
        class_selector=class_selector,
        averager=averager,
        averager_src=averager_src,
    )


class ClassAvgSourcev110(ClassAvgSource):
    """
    Source for denoised 2D images using class average methods.

    Defaults to using Contrast based class selection (on the fly, compressed),
    avoiding neighbors of previous classes,
    and a brute force image alignment.

    Currently this is the most reasonable default for experimental data.
    """

    def __init__(
        self,
        src,
        n_nbor=50,
        num_procs=None,
        classifier=None,
        class_selector=None,
        averager=None,
        averager_src=None,
    ):
        """
        Instantiates ClassAvgSourcev11 with the following parameters.

        :param src: Source used for image classification.
        :param n_nbor: Number of nearest neighbors. Default 50.
        :param num_procs: Number of processors. Use 1 to run serially.
            Default `None` attempts to compute a reasonable value
            based on available cores and memory.
        :param classifier: `Class2D` classifier instance.
            Default `None` creates `RIRClass2D`.
            See code for parameter details.
        :param class_selector: `ClassSelector` instance.
            Default `None` creates `NeighborVarianceWithRepulsionClassSelector`.
        :param averager: `Averager2D` instance.
            Default `None` ceates `BFSRAverager2D` instance.
            See code for parameter details.
        :param averager_src: Optionally explicitly assign source
            to BFSRAverager2D during initialization.
            Raises error when combined with an explicit `averager`
            argument.

        :return: ClassAvgSource instance.
        """
        dtype = src.dtype

        if classifier is None:
            classifier = RIRClass2D(
                src,
                fspca_components=400,
                bispectrum_components=300,  # Compressed Features after last PCA stage.
                n_nbor=n_nbor,
                large_pca_implementation="legacy",
                nn_implementation="sklearn",  # Note this is different than debug
                bispectrum_implementation="legacy",
            )

        if averager is None:
            if averager_src is None:
                averager_src = src

            basis_2d = FFBBasis2D(averager_src.L, dtype=dtype)

            averager = BFSRAverager2D(
                composite_basis=basis_2d,
                src=averager_src,
                num_procs=num_procs,
                dtype=dtype,
            )
        elif averager_src is not None:
            raise RuntimeError(
                "When providing an instantiated `averager`, cannot assign `averager_src`."
            )

        if class_selector is None:
            class_selector = NeighborVarianceWithRepulsionClassSelector()

        super().__init__(
            src=src,
            classifier=classifier,
            class_selector=class_selector,
            averager=averager,
        )



import logging
from collections import OrderedDict

import numpy as np

from aspire.storage import StarFile

logger = logging.getLogger(__name__)


# The metadata_fields dictionary below specifies default data types
# of certain key fields used in the codebase,
# which are originally read from Relion STAR files.
relion_metadata_fields = {
    "_rlnVoltage": float,
    "_rlnDefocusU": float,
    "_rlnDefocusV": float,
    "_rlnDefocusAngle": float,
    "_rlnSphericalAberration": float,
    "_rlnDetectorPixelSize": float,
    "_rlnCtfFigureOfMerit": float,
    "_rlnMagnification": float,
    "_rlnAmplitudeContrast": float,
    "_rlnImageName": str,
    "_rlnOriginalName": str,
    "_rlnCtfImage": str,
    "_rlnCoordinateX": float,
    "_rlnCoordinateY": float,
    "_rlnCoordinateZ": float,
    "_rlnNormCorrection": float,
    "_rlnMicrographName": str,
    "_rlnGroupName": str,
    "_rlnGroupNumber": str,
    "_rlnOriginX": float,
    "_rlnOriginY": float,
    "_rlnAngleRot": float,
    "_rlnAngleTilt": float,
    "_rlnAnglePsi": float,
    "_rlnClassNumber": int,
    "_rlnLogLikeliContribution": float,
    "_rlnRandomSubset": int,
    "_rlnParticleName": str,
    "_rlnOriginalParticleName": str,
    "_rlnNrOfSignificantSamples": float,
    "_rlnNrOfFrames": int,
    "_rlnMaxValueProbDistribution": float,
    "_rlnOpticsGroup": int,
    "_rlnOpticsGroupName": str,
}


def dict_to_relion_types(d):
    """
    Convert STAR file strings to data type for each field in a dict loaded via
    `aspire.storage.StarFile`. Columns without a specified data type are read as
    `dtype=str`.

    :param d: A `StarFile` block represented as a dictionary.
    :return: A dict with types converted where possible.
    """
    column_types = {name: relion_metadata_fields.get(name, str) for name in d}
    # look at values to detect if we're dealing with iterables
    multiple_rows = isinstance(d, dict) and all(
        not isinstance(v, str) and hasattr(v, "__iter__") for v in d.values()
    )
    if multiple_rows:
        retval = dict(zip(d, [np.array(list(map(column_types[k], d[k]))) for k in d]))
    else:
        retval = {k: column_types[k](d[k]) for k in d}
    return retval


class RelionStarFile(StarFile):
    """
    A star file generated by RELION representing particles, micrographs, or movies.
    """

    def __init__(self, filepath):
        super().__init__(filepath, blocks=None)

        # convert dtypes in star file blocks where possible
        self._convert_dtypes()

        # validate star file and detect Relion version (3.0 or >=3.1)
        self._validate_and_detect_version()

    def _validate_and_detect_version(self):
        """
        Based on the number of StarFile blocks, block names, and column names,
        determine whether this star file can be interpreted as coming from RELION.
        If so, determine what version of RELION and what kind of data is represented.
        """

        self.data_block_name = ""
        self.relion_version = ""

        # validate 3.0 STAR file
        if len(self.blocks) == 1:
            data_block = self.get_block_by_index(0)
            columns = list(data_block.keys())
            if not any(
                col in columns
                for col in ["_rlnImageName", "_rlnMicrographName", "_rlnMovieName"]
            ):
                raise ValueError(
                    f"{self.filepath} does not contain Relion data columns."
                )

            self.relion_version = "3.0"
            self.data_block = data_block

        # validate 3.1 STAR file
        if len(self.blocks) == 2:
            # must have an optics block
            if "optics" not in self.blocks.keys():
                raise ValueError(f"{self.filepath} does not contain an optics block.")

            # find type of data
            for name in self.blocks.keys():
                if name in ["particles", "micrographs", "movies"]:
                    self.data_block_name = name
                    break
            if not self.data_block_name:
                raise ValueError(
                    f"{self.filepath} does not contain a block identifying particle, ",
                    "micrograph, or movie data.",
                )

            data_block = self[self.data_block_name]
            # lastly, data block must contain a column identifying the type of data as well
            columns = list(data_block.keys())
            if not any(
                col in columns
                for col in ["_rlnImageName", "_rlnMicrographName", "_rlnMovieName"]
            ):
                raise ValueError(
                    f"{self.filepath} data block does not contain Relion data columns."
                )

            self.relion_version = "3.1"
            self.data_block = data_block
            self.optics_block = self["optics"]

        if not self.relion_version:
            raise ValueError(
                f"Cannot interpret {self.filepath} as a valid RELION STAR file."
            )

    def _convert_dtypes(self):
        """
        For data blocks, convert known Relion columns to sensible types.
        """
        _blocks = OrderedDict()
        for block_name, block in self.blocks.items():
            _blocks[block_name] = dict_to_relion_types(block)
        self.blocks = _blocks

    def get_merged_data_block(self):
        """
        Return the dictionary containing particle/micrograph/movie information for this STAR file.

        :return: A dictionary containing the data from this RELION STAR file.
        """

        if self.relion_version == "3.0":
            # 3.0 star files have the data all in one block, which we return
            return self.data_block

        if self.relion_version == "3.1":
            # merge the parameters in the optics block as new columns in the data block
            # based on the corresponding optics group number (returns a new dataframe)
            data_block = self.data_block.copy()
            # Create a dict from optics_group to individual optics properties
            optics = self.optics_block
            optics_props_by_group_id = dict(
                (k, {_k: optics[_k][i] for _k in optics})
                for i, k in enumerate(optics["_rlnOpticsGroup"])
            )
            # get a NumPy array of optics indices for each row of data
            optics_indices = self.data_block["_rlnOpticsGroup"].astype(int)
            for k in optics:
                data_block[k] = np.array(
                    [
                        optics_props_by_group_id[optic_index][k]
                        for optic_index in optics_indices
                    ]
                )

            return data_block



import logging

import numpy as np

from aspire.basis import Basis
from aspire.basis.basis_utils import (
    d_decay_approx_fun,
    k_operator,
    lgwt,
    t_radial_part_mat,
    t_x_derivative_mat,
    t_x_mat,
)
from aspire.basis.pswf_utils import BNMatrix
from aspire.utils import complex_type

logger = logging.getLogger(__name__)


class PSWFBasis2D(Basis):
    """
    Define a derived class for direct Prolate Spheroidal Wave Function (PSWF) expanding 2D images

    The numerical evaluation for 2D PSWFs at arbitrary points in the unit disk is based on the
    direct method described in the papers as below:
        1) Boris Landa and Yoel Shkolnisky, "Steerable principal components
        for space-frequency localized images", SIAM J. Imag. Sci. 10, 508-534 (2017).
        2) Boris Landa and Yoel Shkolnisky, "Approximation scheme for essentially
        bandlimited and space-concentrated functions on a disk", Appl. Comput.
        Harmon. Anal. 43, 381-403 (2017).
        3) Yoel Shkolnisky, "Prolate spheroidal wave functions on a disc-Integration
        and approximation of two-dimensional bandlimited functions", Appl.
        Comput. Harmon. Anal. 22, 235-256 (2007).
    """

    def __init__(self, size, gamma_trunc=1.0, beta=1.0, dtype=np.float32):
        """
        Initialize an object for 2D PSWF basis expansion using direct method

        :param size: The size of the vectors for which to define the basis
            and the image resolution. May be a 2-tuple or an integer, in which case
            a square basis is assumed. Currently only square images are supported.
        :param gamma_trunc: Truncation parameter of PSWFs, between 0 and 1e6,
            which controls the length of the expansion and the approximation error.
            Smaller values (close to zero) guarantee smaller errors, yet longer
            expansions, and vice-versa. Note: Due to numerical considerations,
            do not exceed 1e6.
        :param beta: Bandlimit ratio relative to the Nyquist rate, between 0 and 1.
            In general, the bandlimit is c = beta*pi*(size[0]//2), therefore for
            the default value beta = 1 there is no oversampling assumed. This
            parameter controls the bandlimit of the PSWFs.
        :param dtype: Internal ndarray datatype.
        """
        if isinstance(size, int):
            size = (size, size)
        self.rcut = size[0] // 2
        self.gmcut = gamma_trunc
        self.beta = beta
        super().__init__(size, dtype=dtype)

        # this basis has complex coefficients
        self.coefficient_dtype = complex_type(self.dtype)

    def _build(self):
        """
        Build internal data structures for the direct 2D PSWF method
        """
        logger.info("Expanding 2D images using direct PSWF method.")

        # initial the whole set of PSWF basis functions based on the bandlimit and eps error.
        self.bandlimit = self.beta * np.pi * self.rcut
        self.d_vec_all, self.alpha_all, self.lengths = self._init_pswf_func2d(
            self.bandlimit, eps=np.spacing(1)
        )

        # generate_the 2D grid and corresponding indices inside the disc.
        self._generate_grid()

        # precompute the basis functions in 2D grids
        self._precomp()

    def _generate_grid(self):
        """
        Generate the 2D sampling grid

        TODO: need to re-implement to use the similar grid function as FB methods.
        """
        if self.nres % 2 == 0:
            x_1d_grid = range(-self.rcut, self.rcut)
        else:
            x_1d_grid = range(-self.rcut, self.rcut + 1)
        x_2d_grid, y_2d_grid = np.meshgrid(x_1d_grid, x_1d_grid)
        r_2d_grid = np.sqrt(np.square(x_2d_grid) + np.square(y_2d_grid))
        points_in_disk = r_2d_grid <= self.rcut
        x = y_2d_grid[points_in_disk]
        y = x_2d_grid[points_in_disk]
        self._r_disk = np.sqrt(np.square(x) + np.square(y)) / self.rcut
        self._theta_disk = np.angle(x + 1j * y)
        self._image_height = len(x_1d_grid)
        self._disk_mask = points_in_disk

    def _precomp(self):
        """
        Precompute PSWF functions on a polar Fourier 2D grid
        """
        self._generate_samples()

        self.non_neg_freq_inds = slice(0, len(self.ang_freqs))

        tmp = np.nonzero(self.ang_freqs == 0)[0]
        self.zero_freq_inds = slice(tmp[0], tmp[-1] + 1)

        tmp = np.nonzero(self.ang_freqs > 0)[0]
        self.pos_freq_inds = slice(tmp[0], tmp[-1] + 1)

    def _generate_samples(self):
        """
        Generate sample points for PSWF functions
        """
        max_ns = []
        a = np.square(float(self.beta * self.rcut) / 2)
        m = 0
        alpha_all = []

        while True:
            alpha = self.alpha_all[m]

            lambda_var = a * np.square(np.absolute(alpha))
            gamma = np.sqrt(np.absolute(lambda_var / (1 - lambda_var)))

            n_end = np.where(gamma <= self.gmcut)[0]

            if len(n_end) != 0:
                n_end = n_end[0]
                if n_end == 0:
                    break
                max_ns.extend([n_end])
                alpha_all.extend(alpha[:n_end])
                m += 1

        self.alpha_nn = np.array(alpha_all).reshape(-1, 1)
        self.max_ns = max_ns

        self.samples = self._evaluate_pswf2d_all(self._r_disk, self._theta_disk, max_ns)
        self.ang_freqs = np.repeat(np.arange(len(max_ns)), max_ns).astype("float")
        self.rad_freqs = np.concatenate([range(1, i + 1) for i in max_ns]).astype(
            "float"
        )
        self.samples = (self.beta / 2.0) * self.samples * self.alpha_nn
        self.samples_conj_transpose = self.samples.conj().transpose()
        # the column dimension of samples_conj_transpose is the number of basis coefficients
        self.count = self.samples_conj_transpose.shape[1]

    def _evaluate_t(self, images):
        """
        Evaluate coefficient vectors in PSWF basis using the direct method

        :param images: coefficient array in the standard 2D coordinate basis
            to be evaluated.
        :return: The evaluation of the coefficient array in the PSWF basis.
        """
        flattened_images = images[:, self._disk_mask]

        return flattened_images @ self.samples_conj_transpose

    def _evaluate(self, coefficients):
        """
        Evaluate coefficients in standard 2D coordinate basis from those in PSWF basis

        :param coeffcients: A coefficient vector (or an array of coefficient
            vectors) in PSWF basis to be evaluated. (n_image, count)
        :return : Image in standard 2D coordinate basis.

        """

        # Handle a single coefficient vector or stack of vectors.
        coefficients = np.atleast_2d(coefficients)
        n_images = coefficients.shape[0]

        angular_is_zero = np.absolute(self.ang_freqs) == 0

        flatten_images = coefficients[:, angular_is_zero] @ self.samples[
            angular_is_zero
        ] + 2.0 * np.real(
            coefficients[:, ~angular_is_zero] @ self.samples[~angular_is_zero]
        )

        images = np.zeros(
            (n_images, self._image_height, self._image_height), dtype=self.dtype
        )
        images[:, self._disk_mask] = np.real(flatten_images)

        return images

    def _init_pswf_func2d(self, c, eps):
        """
        Initialize the whole set of PSWF functions with the input bandlimit and error

        :param c: bandlimit (>0) can be estimated by beta * pi * rcut
        :param eps: error tolerance
        :return:
            alpha_all (list of arrays):
            alpha = alpha_all[i] contains all the eigenvalues for N=i such that
                lambda > eps, where lambda is the normalized  alpha values (i.e.
                lambda is between 0 and 1), given by lambda=sqrt(c*np.absolute(alpha)/(2*pi)).
            d_vec_all (list of 2D lists): the corresponding eigenvectors for alpha_all.
            n_order_length_vec (list of ints): n_order_length_vec[i] = len(alpha_all[i])
        """
        d_vec_all = []
        alpha_all = []
        n_order_length_vec = []

        m = 0
        n = int(np.ceil(2 * c / np.pi))
        r, w = lgwt(n, 0, 1, dtype=self.dtype)

        cons = c / 2 / np.pi
        while True:
            alpha, d_vec, a = self.pswf_func2d(m, n, c, eps, r, w)

            lambda_var = np.sqrt(cons * np.absolute(alpha))

            n_end = np.where(lambda_var <= eps)[0]

            if len(n_end) != 0:
                n_end = n_end[0]
                if n_end == 0:
                    break
                n_order_length_vec.append(n_end)
                alpha_all.append(alpha[:n_end])
                d_vec_all.append(d_vec[:, :n_end])
                m += 1
                n = n_end + 1
            else:
                n *= 2
                r, w = lgwt(n, 0, 1)

        return d_vec_all, alpha_all, n_order_length_vec

    def _evaluate_pswf2d_all(self, r, theta, max_ns):
        """
        Evaluate the numerical values of PSWF functions for all N's, up to given n for each N

        :param r: Radial part to evaluate
        :param theta: Phase part to evaluate
        :param max_ns: List of ints max_ns[i] is max n to to use for N=i, not included.
            If max_ns[i]<1 N=i won't be used
        :return: (sum(max_ns), len(r)) ndarray
            Indices are corresponding to the list (N, n)
            (0, 0),..., (max_ns[0], 0), (0, 1),..., (max_ns[1], 1),... , (0, len(max_ns)-1),
            (max_ns[-1], len(max_ns)-1)
        """
        max_ns_ints = [int(max_n) for max_n in max_ns]
        out_mat = []
        for i, max_n in enumerate(max_ns_ints):
            if max_n < 1:
                continue

            d_vec = self.d_vec_all[i]

            phase_part = np.exp(1j * i * theta) / np.sqrt(2 * np.pi)
            range_array = np.arange(len(d_vec))
            r_radial_part_mat = t_radial_part_mat(r, i, range_array, len(d_vec)).dot(
                d_vec[:, :max_n]
            )

            pswf_n_n_mat = phase_part * r_radial_part_mat.T

            out_mat.extend(pswf_n_n_mat)
        out_mat = np.array(out_mat, dtype=complex_type(self.dtype))
        return out_mat

    def pswf_func2d(self, big_n, n, bandlimit, phi_approximate_error, r, w):
        """
        Calculate the eigenvalues and eigenvectors of PSWF basis functions for all N's and n's

        :param big_n: The integer N in PSWF basis.
        :param n: The integer n in PSWF basis.
        :param bandlimit: The band limit estimated by beta * pi * rcut.
        :param phi_approximate_error: The input approximate error for phi.
        :param r: The Legendre–Gauss quadrature nodes.
        :param w: The Legendre–Gauss quadrature weights.
        :return:
            alpha_n (ndarray): the eigen-values for N.
            d_vec (ndarray): the corresponding eigen-vectors for alpha_n.
            approx_length (int): the number of eigenvalues,len(alpha_n).
        """

        d_vec, approx_length, range_array = self._pswf_2d_minor_computations(
            big_n, n, bandlimit, phi_approximate_error
        )

        t1 = 1 - 2 * np.square(r)
        t2 = np.sqrt(2 * (2 * range_array + big_n + 1))

        phi = t_x_mat(r, big_n, range_array, approx_length).dot(d_vec[:, : (n + 1)])
        phi_derivatives = t_x_derivative_mat(
            t1, t2, r, big_n, range_array, approx_length
        ).dot(d_vec[:, : (n + 1)])

        max_phi_idx = np.argmax(np.absolute(phi[:, 0]))
        max_phi_val = phi[max_phi_idx, 0]
        x_for_calc = r[max_phi_idx]

        right_hand_side_integral = np.einsum(
            "j, j, j ->", w, k_operator(big_n, bandlimit * x_for_calc * r), phi[:, 0]
        )
        lambda_n_1 = right_hand_side_integral / max_phi_val

        temp_calc = r * w
        upper_integral_values = np.einsum(
            "j, ji, ji -> i", temp_calc, phi_derivatives[:, :-1], phi[:, 1:]
        )
        lower_integral_values = np.einsum(
            "j, ji, ji -> i", temp_calc, phi[:, :-1], phi_derivatives[:, 1:]
        )

        lambda_n = np.append(
            np.reshape(lambda_n_1, (1, 1)),
            (
                lambda_n_1
                * np.cumprod(upper_integral_values / lower_integral_values).reshape(
                    (n, 1)
                )
            ),
        )
        alpha_n = lambda_n * 2 * np.pi * (np.power(1j, big_n) / np.sqrt(bandlimit))

        return alpha_n, d_vec, approx_length

    def _pswf_2d_minor_computations(self, big_n, n, bandlimit, phi_approximate_error):
        """
        Approximate the number of n's for fixed N (big_n) and compute the d_vec

        :param big_n: int
        :param n: int
                used for the computation of approx_length
        :param bandlimit: float > 0
        :param phi_approximate_error: float > 0
        :return: d_vec: (approx_length, approx_length) ndarray
                d_vec[:, i] = d^{N, i} defined from eq (18) of paper 3).
            approx_length: int
            range_array: (approx_length,) ndarray
                range_array[i] = i
        """

        first_idx_for_decrease = np.ceil(
            (
                np.sqrt(np.square(2 * n + big_n + 1) + np.square(bandlimit) / 2)
                - (2 * n + big_n + 1)
            )
            / 2
        )

        d_approx = d_decay_approx_fun(big_n, n, bandlimit, first_idx_for_decrease)
        d_decay_index_counter = first_idx_for_decrease

        while d_approx > phi_approximate_error:
            d_decay_index_counter = d_decay_index_counter + 1
            d_approx = d_approx * d_decay_approx_fun(
                big_n, n, bandlimit, d_decay_index_counter
            )
        approx_length = int(n + 1 + d_decay_index_counter)

        d_vec, _ = BNMatrix(big_n, bandlimit, approx_length).get_eig_vectors()

        range_array = np.array(range(approx_length))
        return d_vec, approx_length, range_array



import logging
import os
from warnings import catch_warnings, filterwarnings, simplefilter, warn

import matplotlib.pyplot as plt
import mrcfile
import numpy as np
from PIL import Image as PILImage
from scipy.linalg import lstsq

import aspire.volume
from aspire.nufft import anufft
from aspire.numeric import fft, xp
from aspire.utils import FourierRingCorrelation, anorm, crop_pad_2d, grid_2d

logger = logging.getLogger(__name__)


def normalize_bg(imgs, bg_radius=1.0, do_ramp=True):
    """
    Normalize backgrounds and apply to a stack of images

    :param imgs: A stack of images in N-by-L-by-L array
    :param bg_radius: Radius cutoff to be considered as background (in image size)
    :param do_ramp: When it is `True`, fit a ramping background to the data
            and subtract. Namely perform normalization based on values from each image.
            Otherwise, a constant background level from all images is used.
    :return: The modified images
    """
    if imgs.ndim > 3:
        raise NotImplementedError(
            "`normalize_bg` is currently limited to 1D image stacks."
        )

    L = imgs.shape[-1]
    input_dtype = imgs.dtype
    grid = grid_2d(L, indexing="yx", dtype=input_dtype)
    mask = grid["r"] > bg_radius

    if do_ramp:
        # Create matrices and reshape the background mask
        # for fitting a ramping background
        ramp_mask = np.vstack(
            (
                grid["x"][mask].flatten(),
                grid["y"][mask].flatten(),
                np.ones(grid["y"][mask].flatten().size, dtype=input_dtype),
            )
        ).T
        ramp_all = np.vstack(
            (
                grid["x"].flatten(),
                grid["y"].flatten(),
                np.ones(L * L, dtype=input_dtype),
            )
        ).T
        mask_reshape = mask.reshape((L * L))
        imgs = imgs.reshape((-1, L * L))

        # Fit a ramping background and apply to images
        coeff = lstsq(ramp_mask, imgs[:, mask_reshape].T)[0]  # RCOPT
        imgs = imgs - (ramp_all @ coeff).T  # RCOPT
        imgs = imgs.reshape((-1, L, L))

    # Apply mask images and calculate mean and std values of background
    imgs_masked = imgs * mask
    denominator = np.sum(mask)
    first_moment = np.sum(imgs_masked, axis=(1, 2)) / denominator
    second_moment = np.sum(imgs_masked**2, axis=(1, 2)) / denominator
    mean = first_moment.reshape(-1, 1, 1)
    variance = second_moment.reshape(-1, 1, 1) - mean**2
    std = np.sqrt(variance)

    return (imgs - mean) / std


def load_mrc(filepath):
    """
    Load raw data from `.mrc` into an array.

    :param filepath: File path (string).
    :return: numpy array of image data.
    """

    # mrcfile tends to yield many warnings about EMPIAR datasets being corrupt
    # These warnings generally seem benign, and the message could be clearer
    # The following code handles the warnings via ASPIRE's logger
    with catch_warnings(record=True) as ws:
        # Cause all warnings to always be triggered in this context
        simplefilter("always")

        with mrcfile.open(filepath, mode="r", permissive=True) as mrc:
            im = mrc.data

        # Log each mrcfile warning to debug log, noting the associated file
        for w in ws:
            logger.debug(
                "In `Image.load` mrcfile.open reports corruption for"
                f" {filepath} warning: {w.message}"
            )

        # Log a single warning to user
        # Give context and note assocated filepath
        if len(ws) > 0:
            logger.warning(
                f"Image.load of {filepath} reporting {len(ws)} corruptions."
                " Most likely this is a problem with the header contents."
                " Details written to debug log."
                f" Will attempt to continue processing {filepath}"
            )

    return im


def load_tiff(filepath):
    """
    Load raw data from `.tiff` into an array.

    :param filepath: File path (string).
    :return: numpy array of image data.
    """

    # Use PIL to open `filepath` and cast to numpy array.
    return np.array(PILImage.open(filepath))


class Image:
    # Map file extensions to their respective readers
    extensions = {
        ".mrc": load_mrc,
        ".tif": load_tiff,
        ".tiff": load_tiff,
    }

    def __init__(self, data, dtype=None):
        """
        A stack of one or more images.

        This is a wrapper of numpy.ndarray which provides methods
        for common processing tasks.

        The stack can be multidimensional with `n_images` equal
        to the product of the stack dimensions.  Singletons will be
        expanded into a stack with one entry.

        The last two axes represent the image size,
        and are checked to be square.

        :param data: Numpy array containing image data with shape
            `(..., resolution, resolution)`.
        :param dtype: Optionally cast `data` to this dtype.
            Defaults to `data.dtype`.

        :return: Image instance holding `data`.
        """

        if not isinstance(data, np.ndarray):
            raise ValueError("Image should be instantiated with an ndarray")

        if data.ndim < 2:
            raise ValueError(
                "Image data should be ndarray with shape (N1...)xLxL or LxL."
            )
        elif data.ndim == 2:
            data = np.expand_dims(data, axis=0)

        if dtype is None:
            self.dtype = data.dtype
        else:
            self.dtype = np.dtype(dtype)

        if not data.shape[-1] == data.shape[-2]:
            raise ValueError("Only square ndarrays are supported.")

        self._data = data.astype(self.dtype, copy=False)
        self.ndim = self._data.ndim
        self.shape = self._data.shape
        self.stack_ndim = self._data.ndim - 2
        self.stack_shape = self._data.shape[:-2]
        self.n_images = np.prod(self.stack_shape)
        self.resolution = self._data.shape[-1]

        # Numpy interop
        # https://numpy.org/devdocs/user/basics.interoperability.html#the-array-interface-protocol
        self.__array_interface__ = self._data.__array_interface__
        self.__array__ = self._data

    @property
    def res(self):
        warn(
            "`Image.res` will be deprecated in favor or Image.resolution in an upcoming release.",
            DeprecationWarning,
        )
        return self.resolution

    def _check_key_dims(self, key):
        if isinstance(key, tuple) and (len(key) > self._data.ndim):
            raise ValueError(
                f"Image stack_dim is {self.stack_ndim}, slice length must be =< {self.ndim}"
            )

    def __getitem__(self, key):
        self._check_key_dims(key)
        return self.__class__(self._data[key])

    def __setitem__(self, key, value):
        self._check_key_dims(key)
        self._data[key] = value

    def stack_reshape(self, *args):
        """
        Reshape the stack axis.

        :*args: Integer(s) or tuple describing the intended shape.

        :returns: Image instance
        """

        # If we're passed a tuple, use that
        if len(args) == 1 and isinstance(args[0], tuple):
            shape = args[0]
        else:
            # Otherwise use the variadic args
            shape = args

        # Sanity check the size
        if shape != (-1,) and np.prod(shape) != self.n_images:
            raise ValueError(
                f"Number of images {self.n_images} cannot be reshaped to {shape}."
            )

        return self.__class__(self._data.reshape(*shape, *self._data.shape[-2:]))

    def __add__(self, other):
        if isinstance(other, Image):
            other = other._data

        return self.__class__(self._data + other)

    def __sub__(self, other):
        if isinstance(other, Image):
            other = other._data

        return self.__class__(self._data - other)

    def __mul__(self, other):
        if isinstance(other, Image):
            other = other._data

        return self.__class__(self._data * other)

    def __neg__(self):
        return self.__class__(-self._data)

    def sqrt(self):
        return self.__class__(np.sqrt(self._data))

    @property
    def T(self):
        """
        Abbreviation for transpose.

        :return: Image instance.
        """
        return self.transpose()

    def transpose(self):
        """
        Returns a new Image instance with image data axes transposed.

        :return: Image instance.
        """
        original_stack_shape = self.stack_shape

        im = self.stack_reshape(-1)
        imt = np.transpose(im._data, (0, -1, -2))

        return self.__class__(imt).stack_reshape(original_stack_shape)

    def flip(self, axis=-2):
        """
        Flip image stack data along axis using numpy.flip().

        :param axis: Optionally specify axis as integer or tuple.
            Defaults to axis=-2.

        :return: Image instance.
        """
        # Convert integer to tuple, so we can always loop.
        if isinstance(axis, int):
            axis = (axis,)

        # Check we are not attempting to flip any stack axis.
        for ax in axis:
            ax = ax % self.ndim  # modulo [0, ndim)
            if ax < self.stack_ndim:
                raise ValueError(
                    f"Cannot flip axis {ax}: stack axis. Did you mean {ax-3}?"
                )

        return self.__class__(np.flip(self._data, axis))

    def __repr__(self):
        msg = f"{self.n_images} {self.dtype} images arranged as a {self.stack_shape} stack"
        msg += f" each of size {self.resolution}x{self.resolution}."
        return msg

    def asnumpy(self):
        """
        Return image data as a (<stack>, resolution, resolution)
        read-only array view.

        :return: read-only ndarray view
        """

        view = self._data.view()
        view.flags.writeable = False
        return view

    def copy(self):
        return self.__class__(self._data.copy())

    def shift(self, shifts):
        """
        Translate image by shifts. This method returns a new Image.

        :param shifts: An array of size n-by-2 specifying the shifts in pixels.
            Alternatively, it can be a column vector of length 2, in which case
            the same shifts is applied to each image.
        :return: The Image translated by the shifts, with periodic boundaries.
        """
        if shifts.ndim == 1:
            shifts = shifts[np.newaxis, :]

        n_shifts = shifts.shape[0]

        if not shifts.shape[1] == 2:
            raise ValueError("Input shifts must be of shape (n_images, 2) or (1, 2).")
        if not n_shifts == 1 and not n_shifts == self.n_images:
            raise ValueError(
                "The number of shifts must be 1 or equal to self.n_images."
            )

        return self._im_translate(shifts)

    def downsample(self, ds_res):
        """
        Downsample Image to a specific resolution. This method returns a new Image.

        :param ds_res: int - new resolution, should be <= the current resolution
            of this Image
        :return: The downsampled Image object.
        """

        original_stack_shape = self.stack_shape
        im = self.stack_reshape(-1)

        # compute FT with centered 0-frequency
        fx = fft.centered_fft2(im._data)
        # crop 2D Fourier transform for each image
        crop_fx = np.array([crop_pad_2d(fx[i], ds_res) for i in range(self.n_images)])
        # take back to real space, discard complex part, and scale
        out = np.real(fft.centered_ifft2(crop_fx)) * (
            ds_res**2 / self.resolution**2
        )

        return self.__class__(out).stack_reshape(original_stack_shape)

    def filter(self, filter):
        """
        Apply a `Filter` object to the Image and returns a new Image.

        :param filter: An object of type `Filter`.
        :return: A new filtered `Image` object.
        """
        original_stack_shape = self.stack_shape

        im = self.stack_reshape(-1)

        filter_values = filter.evaluate_grid(self.resolution)

        im_f = xp.asnumpy(fft.centered_fft2(xp.asarray(im._data)))

        # TODO: why are these different? Doesn't the broadcast work?
        if im_f.ndim > filter_values.ndim:
            im_f *= filter_values
        else:
            im_f = filter_values * im_f
        im = xp.asnumpy(fft.centered_ifft2(xp.asarray(im_f)))
        im = np.real(im)

        return self.__class__(im).stack_reshape(original_stack_shape)

    def rotate(self):
        raise NotImplementedError

    def save(self, mrcs_filepath, overwrite=False):
        if self.stack_ndim > 1:
            raise NotImplementedError("`save` is currently limited to 1D image stacks.")

        with mrcfile.new(mrcs_filepath, overwrite=overwrite) as mrc:
            # original input format (the image index first)
            mrc.set_data(self._data.astype(np.float32))

    @staticmethod
    def load(filepath, dtype=None):
        """
        Load raw data from supported files.

        Currently MRC and TIFF are supported.

        :param filepath: File path (string).
        :param dtype: Optionally force cast to `dtype`.
             Default dtype is inferred from the file contents.
        :return: numpy array of image data.
        """

        # Get the file extension
        ext = os.path.splitext(filepath)[1]

        # On unsupported extension, raise with suggested file types
        if ext not in Image.extensions:
            raise RuntimeError(
                f"Attempting to open unsupported file extension '{ext}', try {list(Image.extensions.keys())}."
            )

        # Call the appropriate file reader
        im = Image.extensions[ext](filepath)

        # Attempt casting when user provides dtype
        if dtype is not None:
            im = im.astype(dtype, copy=False)

        # Return as Image instance
        return Image(im)

    def _im_translate(self, shifts):
        """
        Translate image by shifts
        :param im: An array of size n-by-L-by-L containing images to be translated.
        :param shifts: An array of size n-by-2 specifying the shifts in pixels.
            Alternatively, it can be a row vector of length 2, in which case the same shifts is applied to each image.
        :return: The images translated by the shifts, with periodic boundaries.
        """

        # Note original stack shape and flatten stack
        stack_shape = self.stack_shape
        im = self.stack_reshape(-1)._data

        if shifts.ndim == 1:
            shifts = shifts[np.newaxis, :]
        n_shifts = shifts.shape[0]

        assert shifts.shape[-1] == 2, "shifts must be nx2"

        assert (
            n_shifts == 1 or n_shifts == self.n_images
        ), "number of shifts must be 1 or match the number of images"
        # Cast shifts to this instance's internal dtype
        shifts = shifts.astype(self.dtype)

        L = self.resolution
        im_f = xp.asnumpy(fft.fft2(xp.asarray(im)))
        grid_shifted = fft.ifftshift(
            xp.asarray(np.ceil(np.arange(-L / 2, L / 2, dtype=self.dtype)))
        )
        grid_1d = xp.asnumpy(grid_shifted) * 2 * np.pi / L
        om_x, om_y = np.meshgrid(grid_1d, grid_1d, indexing="ij")

        phase_shifts_x = -shifts[:, 0].reshape((n_shifts, 1, 1))
        phase_shifts_y = -shifts[:, 1].reshape((n_shifts, 1, 1))

        phase_shifts = (
            om_x[np.newaxis, :, :] * phase_shifts_x
            + om_y[np.newaxis, :, :] * phase_shifts_y
        )
        mult_f = np.exp(-1j * phase_shifts)
        im_translated_f = im_f * mult_f
        im_translated = xp.asnumpy(fft.ifft2(xp.asarray(im_translated_f)))
        im_translated = np.real(im_translated)

        # Reshape to stack shape
        return self.__class__(im_translated).stack_reshape(stack_shape)

    def norm(self):
        return anorm(self._data)

    @property
    def size(self):
        # probably not needed, transition
        return np.size(self._data)

    def backproject(self, rot_matrices):
        """
        Backproject images along rotation
        :param im: An Image (stack) to backproject.
        :param rot_matrices: An n-by-3-by-3 array of rotation matrices \
        corresponding to viewing directions.

        :return: Volume instance corresonding to the backprojected images.
        """

        if self.stack_ndim > 1:
            raise NotImplementedError(
                "`Backprojection` is currently limited to 1D image stacks."
            )

        L = self.resolution

        assert (
            self.n_images == rot_matrices.shape[0]
        ), "Number of rotation matrices must match the number of images"

        # TODO: rotated_grids might as well give us correctly shaped array in the first place
        pts_rot = aspire.volume.rotated_grids(L, rot_matrices).astype(
            self.dtype, copy=False
        )
        pts_rot = pts_rot.reshape((3, -1))

        im_f = xp.asnumpy(fft.centered_fft2(xp.asarray(self._data))) / (L**2)
        if L % 2 == 0:
            im_f[:, 0, :] = 0
            im_f[:, :, 0] = 0

        im_f = im_f.flatten()

        vol = anufft(im_f, pts_rot[::-1], (L, L, L), real=True) / L

        return aspire.volume.Volume(vol)

    def show(self, columns=5, figsize=(20, 10), colorbar=True):
        """
        Plotting Utility Function.

        :param columns: Number of columns in a row of plots.
        :param figsize: Figure size in inches, consult `matplotlib.figure`.
        :param colorbar: Optionally plot colorbar to show scale.
            Defaults to True. Accepts `bool` or `dictionary`,
            where the dictionary is passed to `matplotlib.pyplot.colorbar`.
        """

        if self.stack_ndim > 1:
            raise NotImplementedError("`show` is currently limited to 1D image stacks.")

        # We never need more columns than images.
        columns = min(columns, self.n_images)
        rows = (self.n_images + columns - 1) // columns  # ceiling divide.

        # Create an empty colorbar options dictionary as needed.
        colorbar_opts = colorbar if isinstance(colorbar, dict) else dict()

        # Create a context manager for altering warnings
        with catch_warnings():
            # Filter off specific warning.
            # sphinx-gallery overrides to `agg` backend, but doesn't handle warning.
            filterwarnings(
                "ignore",
                category=UserWarning,
                message="Matplotlib is currently using agg, which is a"
                " non-GUI backend, so cannot show the figure.",
            )

            plt.figure(figsize=figsize)
            for i, im in enumerate(self.asnumpy()):
                plt.subplot(rows, columns, i + 1)
                plt.imshow(im, cmap="gray")
                if colorbar:
                    plt.colorbar(**colorbar_opts)

            plt.show()

    def frc(self, other, cutoff, pixel_size=None, method="fft", plot=False):
        r"""
        Compute the Fourier ring correlation between two images.

        Images are assumed to be well aligned.

        The FRC is defined as:

        .. math::

           c(i) = \frac{ \operatorname{Re}( \sum_i{ \mathcal{F}_1(i) * {\mathcal{F}^{*}_2(i) } } ) }{\
             \sqrt{ \sum_i { | \mathcal{F}_1(i) |^2 } * \sum_i{| \mathcal{F}^{*}_2}(i) |^2 } }

        :param other: `Image` instance to compare.
        :param cutoff: Cutoff value, traditionally `.143`.
        :param pixel_size: Pixel size in angstrom.  Default `None`
            implies unit in pixels, equivalent to pixel_size=1.
        :param method: Selects either 'fft' (on cartesian grid),
            or 'nufft' (on polar grid). Defaults to 'fft'.
        :param plot: Optionally plot to screen or file.
            Defaults to `False`.  `True` plots to screen.
            Passing a filepath as a string will attempt to save to file.

        :return: tuple(estimated_resolution,  FRC),
            where `estimated_resolution` is in angstrom
            and FRC is a Numpy array of correlations.
        """

        if not isinstance(other, Image):
            raise TypeError(
                f"`other` image must be an `Image` instance, received {type(other)}"
            )

        frc = FourierRingCorrelation(
            a=self.asnumpy(),
            b=other.asnumpy(),
            pixel_size=pixel_size,
            method=method,
        )

        if plot is True:
            frc.plot(cutoff=cutoff)
        elif plot:
            frc.plot(cutoff=cutoff, save_to_file=plot)

        return frc.analyze_correlations(cutoff), frc.correlations


class CartesianImage(Image):
    def expand(self, basis):
        return BasisImage(basis)


class PolarImage(Image):
    def expand(self, basis):
        return BasisImage(basis)


class BispecImage(Image):
    def expand(self, basis):
        return BasisImage(basis)


class BasisImage(Image):
    def __init__(self, basis):
        self.basis = basis

    def evaluate(self):
        return CartesianImage()


class FBBasisImage(BasisImage):
    pass



"""
Define a Rotation Class for customized rotation operations used by ASPIRE.
"""

import warnings

import numpy as np
from numpy.linalg import norm
from scipy.linalg import svd
from scipy.spatial.transform import Rotation as sp_rot

from aspire.utils.random import Random


class Rotation:
    def __init__(self, matrices, gimble_lock_warnings=True):
        """
         Initialize a Rotation object

        :param matrices: Rotation matrices to initialize Rotation object.
        :param gimble_lock_warnings: Optionally show gimble lock warnings.
             Defaults to `True`, setting to `False` filters warnings.
        """
        if matrices.ndim == 2:
            matrices = matrices.reshape((1, 3, 3))
        assert matrices.ndim == 3 and matrices.shape[-2:] == (
            3,
            3,
        ), f"Bad rotation matrix shape: {matrices.shape}"
        self._matrices = matrices
        self._seq_order = "ZYZ"

        self.gimble_lock_warnings = bool(gimble_lock_warnings)

    def __str__(self):
        """
        String representation.
        """
        return f"Rotation stack consisting of {len(self)} elements of {self.dtype} type"

    def __len__(self):
        return self._matrices.shape[0]

    def __getitem__(self, item):
        return self._matrices[item]

    def __setitem__(self, key, value):
        self._matrices[key] = value

    def __mul__(self, other):
        output = self._matrices @ other.matrices
        return Rotation(output)

    @property
    def dtype(self):
        """
        :return: dtype of Rotation matrices
        """
        return self._matrices.dtype

    @property
    def matrices(self):
        """
        :return: Rotation matrices as a n x 3 x 3 array
        """
        return self._matrices

    @property
    def angles(self):
        """
        :return: Rotation matrices as a n x 3 array
        """
        rotations = sp_rot.from_matrix(self._matrices.astype(self.dtype))

        # Filter Gimbal lock warnings.
        with warnings.catch_warnings():
            msg = "Gimbal lock detected*"

            disposition = "once"
            if not self.gimble_lock_warnings:
                disposition = "ignore"

            warnings.filterwarnings(disposition, message=msg, category=UserWarning)

            euler_angles = rotations.as_euler(self._seq_order, degrees=False).astype(
                self.dtype
            )

        return euler_angles

    def invert(self):
        """
        Apply transpose operation to all rotation matrices

        :return: The set of transposed matrices
        """
        return Rotation(
            matrices=np.transpose(self._matrices, axes=(0, 2, 1)),
        )

    def find_registration(self, rots_ref):
        """
        Register estimated orientations to reference ones.

        Finds the orthogonal transformation that best aligns the estimated rotations
        to the reference rotations.

        :param rots_ref: The reference Rotation object to which we would like to align
            with data matrices in the form of a n-by-3-by-3 array.
        :return: o_mat, optimal orthogonal 3x3 matrix to align the two sets;
                flag, flag==1 then J conjugacy is required and 0 is not.
        """
        rots = self._matrices
        rots_ref = rots_ref.matrices.astype(self.dtype)
        assert (
            rots.shape == rots_ref.shape
        ), "Two sets of rotations must have same dimensions."
        K = rots.shape[0]

        # Reflection matrix
        J = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])

        Q1 = np.zeros((3, 3), dtype=rots.dtype)
        Q2 = np.zeros((3, 3), dtype=rots.dtype)

        for k in range(K):
            R = rots[k, :, :]
            Rref = rots_ref[k, :, :]
            Q1 = Q1 + R @ Rref.T
            Q2 = Q2 + (J @ R @ J) @ Rref.T

        # Compute the two possible orthogonal matrices which register the
        # estimated rotations to the true ones.
        Q1 = Q1 / K
        Q2 = Q2 / K

        # We are registering one set of rotations (the estimated ones) to
        # another set of rotations (the true ones). Thus, the transformation
        # matrix between the two sets of rotations should be orthogonal. This
        # matrix is either Q1 if we recover the non-reflected solution, or Q2,
        # if we got the reflected one. In any case, one of them should be
        # orthogonal.

        err1 = norm(Q1 @ Q1.T - np.eye(3), ord="fro")
        err2 = norm(Q2 @ Q2.T - np.eye(3), ord="fro")

        # In any case, enforce the registering matrix O to be a rotation.
        if err1 < err2:
            # Use Q1 as the registering matrix
            U, _, V = svd(Q1)
            flag = 0
        else:
            # Use Q2 as the registering matrix
            U, _, V = svd(Q2)
            flag = 1

        Q_mat = U @ V

        return Q_mat, flag

    def apply_registration(self, Q_mat, flag):
        """
        Get aligned Rotation object to reference ones.

        Calculated aligned rotation matrices from the orthogonal transformation
        that best aligns the estimated rotations to the reference rotations.

        :param Q_mat:  optimal orthogonal 3x3 transformation matrix
        :param flag:  flag==1 then J conjugacy is required and 0 is not
        :return: regrot, aligned Rotation object
        """
        rots = self._matrices
        K = rots.shape[0]

        # Reflection matrix
        J = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])

        regrot = np.zeros_like(rots)
        for k in range(K):
            R = rots[k, :, :]
            if flag == 1:
                R = J @ R @ J
            regrot[k, :, :] = Q_mat.T @ R
        aligned_rots = Rotation(regrot)
        return aligned_rots

    def register(self, rots_ref):
        """
         Estimate global orientation and return an aligned Rotation object.

        :param rots_ref: The reference Rotation object to which we would like
            to align with data matrices in the form of a n-by-3-by-3 array.
        :return: an aligned Rotation object
        """
        Q_mat, flag = self.find_registration(rots_ref)
        return self.apply_registration(Q_mat, flag)

    def mse(self, rots_ref):
        """
        Calculate MSE between the estimated orientations to reference ones.

        :param rots_reg: The estimated Rotation object after alignment
             with data matrices in the form of a n-by-3-by-3 array.
        :param rots_ref: The reference Rotation object.
        :return: The MSE value between two sets of rotations.
        """
        aligned_rots = self.register(rots_ref)
        rots_reg = aligned_rots.matrices
        rots_ref = rots_ref.matrices
        assert (
            rots_reg.shape == rots_ref.shape
        ), "Two sets of rotations must have same dimensions."
        K = rots_reg.shape[0]

        diff = np.zeros(K)
        mse = 0
        for k in range(K):
            diff[k] = norm(rots_reg[k, :, :] - rots_ref[k, :, :], ord="fro")
            mse += diff[k] ** 2
        mse = mse / K
        return mse

    def common_lines(self, i, j, ell):
        """
        Compute the common line induced by rotation matrices i and j.

        :param i: The index of first rotation matrix of 3-by-3 array.
        :param j: The index of second rotation matrix of 3-by-3 array.
        :param ell: The total number of common lines.
        :return: The common line indices for both first and second rotations.
        """
        r1 = self._matrices[i]
        r2 = self._matrices[j]
        ut = np.dot(r2, r1.T)
        alpha_ij = np.arctan2(ut[2, 0], -ut[2, 1]) + np.pi
        alpha_ji = np.arctan2(ut[0, 2], -ut[1, 2]) + np.pi

        ell_ij = alpha_ij * ell / (2 * np.pi)
        ell_ji = alpha_ji * ell / (2 * np.pi)

        ell_ij = int(np.mod(np.round(ell_ij), ell))
        ell_ji = int(np.mod(np.round(ell_ji), ell))

        return ell_ij, ell_ji

    @staticmethod
    def from_euler(values, dtype=None):
        """
        build rotation object from Euler angles in radians

        :param values: Rotation angles in radians, as a n x 3 array
        :param dtype: Optional output dtype.  Infers from `values` by default when possible,
            otherwise defaults to `float32`.

        :return: new Rotation object
        """
        dtype = dtype or getattr(values, "dtype", np.float32)
        rotations = sp_rot.from_euler("ZYZ", values, degrees=False)
        matrices = rotations.as_matrix().astype(dtype, copy=False)
        return Rotation(matrices)

    @staticmethod
    def about_axis(axis, angles, dtype=None, gimble_lock_warnings=True):
        """
        Build rotation object from axis and angles of rotation.

        :param axis: A string denoting the axis of rotation. "x", "y", or "z".
        :param angles: Rotation angles in radians. `angles` can be a single value,
            or an array of shape (N,) or (N,1).
        :param dtype: Optional output dtype.  Infers from `angles` by default when possible,
            otherwise defaults to `float32`.
        :param gimble_lock_warnings: Optionally show gimble lock warnings.
             Defaults to `True`, setting to `False` filters warnings.

        :return: Rotation object
        """
        dtype = dtype or getattr(angles, "dtype", np.float32)
        axes = ["x", "y", "z"]
        if axis.lower() not in axes:
            raise ValueError("`axis` must be 'x', 'y', or 'z'.")

        angles = np.asarray(angles, dtype=dtype)
        if angles.ndim > 2:
            raise ValueError(
                f"`angles` must be float, 1D array, or 2D array. Got shape {angles.shape}."
            )
        elif angles.ndim == 2 and angles.shape[-1] != 1:
            raise ValueError(
                f"Expected `angles` to have shape (N,1), got {angles.shape}"
            )

        rotation = sp_rot.from_euler(axis, angles, degrees=False)
        matrix = rotation.as_matrix().astype(dtype)
        rot = Rotation(matrix, gimble_lock_warnings=gimble_lock_warnings)

        return rot

    def as_rotvec(self):
        """
        Return this `Rotation` object as a rotation vector.

        A rotation vector is a 3D vector which is co-directional to
        the axis of rotation and whose norm gives the angle of
        rotation in radians. The angle of rotation is
        counter-clockwise about the axis.

        See `from_rotvec`.

        :return: (..., 3) array
        """
        return sp_rot.from_matrix(self.matrices).as_rotvec()

    @staticmethod
    def from_rotvec(vec, dtype=None):
        """
        Build a Rotation object from rotation vectors. A rotation vector is a
        3D vector which is co-directional to the axis of rotation and whose norm
        gives the angle of rotation in radians. The angle of rotation is counter-clockwise
        about the axis.

        :param vec: array_like, shape (N, 3) or (3,)
        :param dtype: Optional output dtype.  Infers from `vec` by default.
        :return: Rotation object
        """
        dtype = dtype or vec.dtype
        rots = sp_rot.from_rotvec(vec)
        matrices = rots.as_matrix().astype(dtype, copy=False)
        return Rotation(matrices)

    @staticmethod
    def from_matrix(values, dtype=None):
        """
        build rotation object from rotational matrices

        :param dtype: Optional output dtype.  Infers from `values` by default.
        :param values: Rotation matrices, as a n x 3 x 3 array
        :return: new Rotation object
        """
        dtype = dtype or values.dtype
        return Rotation(values.astype(dtype, copy=False))

    @staticmethod
    def generate_random_rotations(
        n,
        seed=None,
        dtype=np.float32,
    ):
        """
        Generate Rotation object with random 3D rotation matrices

        :param n: The number of rotation matrices to generate
        :param seed: Random integer seed to use. If None,
            the current random state is used.
        :param dtype:  data type for rotational angles and matrices
        :return: A new Rotation object
        """
        # Generate random rotation angles, in radians
        with Random(seed):
            angles = np.column_stack(
                (
                    np.random.random(n) * 2 * np.pi,
                    np.arccos(2 * np.random.random(n) - 1),
                    np.random.random(n) * 2 * np.pi,
                )
            ).astype(dtype=dtype)

        return Rotation.from_euler(angles, dtype=dtype)

    @staticmethod
    def angle_dist(r1, r2, dtype=None):
        """
        Find the angular distance between two sets of rotation matrices. We first compute
        the rotation between the two sets of rotation matrices, r = r1 @ r2.T. Then using
        the axis-angle representation of r we find the angle between r1 and r2. Computations
        will broadcast in the case of a singleton rotation and a set of rotations.

        :param r1: An nx3x3 array of rotation matrices, a singleton rotation,
            or a `Rotation` object.
        :param r2: An nx3x3 array of rotation matrices, a singleton rotation,
            or a `Rotation` object.
        :param dtype: Computation datatype. Default `None` infers from `r1`.

        :return: The element-wise angular distance between r1 and r2 in radians.
        """
        if isinstance(r1, Rotation):
            r1 = r1.matrices
        if isinstance(r2, Rotation):
            r2 = r2.matrices

        # Handle singletons.
        r1 = r1.reshape(-1, 3, 3)
        r2 = r2.reshape(-1, 3, 3)

        # Check that shapes are compatible.
        if (r1.shape[0] != 1) and (r2.shape[0] != 1):
            if r1.shape[0] != r2.shape[0]:
                raise ValueError(r"r1 and r2 are not broadcastable.*")

        dtype = np.dtype(dtype or r1.dtype)

        r2_T = np.transpose(r2, axes=(0, 2, 1))
        r = r1 @ r2_T
        tr_r = np.trace(r, axis1=1, axis2=2, dtype=dtype)

        dist = np.zeros(max(len(r1), len(r2)), dtype=dtype)
        non_zero_dist_ind = np.where(abs(tr_r - 3.0) > np.finfo(tr_r.dtype).resolution)

        theta = (tr_r[non_zero_dist_ind] - 1) / 2
        theta = np.maximum(np.minimum(theta, 1), -1)  # Clamp theta in [-1,1]
        dist[non_zero_dist_ind] = np.arccos(theta, dtype=dtype)
        return dist

    @staticmethod
    def mean_angular_distance(rots_1, rots_2, dtype=None):
        """
        Find the mean angular distance between two sets of rotation matrices. Computations
        will broadcast in the case of a singleton rotations and a set of rotations.

        :param rots_1: An nx3x3 array of rotation matrices, a singleton rotation,
            or a `Rotation` object.
        :param rots_2: An nx3x3 array of rotation matrices, a singleton rotation,
            or a `Rotation` object.
        :param dtype: Data type for computation. Default infers dtype from `rots_1`.

        :return: The mean angular distance between rotations in radians.
        """

        dtype = np.dtype(dtype or rots_1.dtype)

        # Shapes will be checked by `angle_dist`.
        ang_dist = Rotation.angle_dist(rots_1, rots_2, dtype=dtype)

        return np.mean(ang_dist)



# isort: off
from .registry import file_to_method_map, registry, registry_urls

# isort: on
from .data_fetcher import (
    available_downloads,
    emdb_2484,
    emdb_2660,
    emdb_2824,
    emdb_2984,
    emdb_3645,
    emdb_4905,
    emdb_5778,
    emdb_6287,
    emdb_6458,
    emdb_8012,
    emdb_8511,
    emdb_10835,
    emdb_14621,
    remove_downloads,
)



from .image import (
    BasisImage,
    BispecImage,
    CartesianImage,
    FBBasisImage,
    Image,
    PolarImage,
    load_mrc,
    load_tiff,
    normalize_bg,
)
from .image_stacker import (
    ImageStacker,
    MeanImageStacker,
    MedianImageStacker,
    SigmaRejectionImageStacker,
    WinsorizedImageStacker,
)



import logging

import click

# Overrides click.option with ASPIRE global defaults, see ./__init__.py
import aspire.commands._config  # noqa: F401
from aspire import ctf
from aspire.commands import log_level_option
from aspire.utils.logging import setConsoleLoggingLevel

logger = logging.getLogger("aspire")


@click.command()
@click.option("--data_folder", default=None, help="Path to mrc or mrcs files")
@click.option(
    "--pixel_size", default=1, type=float, help="Pixel size in \u212b (angstrom)"
)
@click.option("--cs", default=2.0, type=float, help="Spherical aberration")
@click.option(
    "--amplitude_contrast", default=0.07, type=float, help="Amplitude contrast"
)
@click.option(
    "--voltage",
    default=300,
    type=float,
    help="Electron microscope Voltage in kilovolts (kV)",
)
@click.option(
    "--num_tapers",
    default=2,
    type=int,
    help="Number of tapers to apply in PSD estimation",
)
@click.option(
    "--psd_size",
    default=512,
    type=int,
    help="Size of blocks for use in Power Spectrum estimation",
)
@click.option(
    "--g_min",
    default=30,
    type=float,
    help="Inverse of minimum resolution for Power Spectrum Distribution",
)
@click.option(
    "--g_max", default=5, type=float, help="Inverse of maximum resolution for PSD"
)
@click.option(
    "--dtype",
    default="float32",
    help="NumPy dtype to use in computation.  Example: 'float32' or 'float64'.",
)
@click.option(
    "--output_dir",
    default="results",
    help="Path to output files, defaults to './results'",
)
@log_level_option
def estimate_ctf(
    data_folder,
    pixel_size,
    cs,
    amplitude_contrast,
    voltage,
    num_tapers,
    psd_size,
    g_min,
    g_max,
    output_dir,
    dtype,
    loglevel,
):
    """
    Given parameters estimates CTF from experimental data
    and returns CTF as a mrc file.

    This is a Click command line interface wrapper for
    the aspire.ctf module.
    """

    # Set desired logging option for the command line
    setConsoleLoggingLevel(loglevel)

    return ctf.estimate_ctf(
        data_folder,
        pixel_size,
        cs,
        amplitude_contrast,
        voltage,
        num_tapers,
        psd_size,
        g_min,
        g_max,
        output_dir,
        dtype,
    )



import logging

import numpy as np

from aspire.image import Image
from aspire.nufft import nufft
from aspire.utils import complex_type

logger = logging.getLogger(__name__)


class PolarFT:
    """
    Define a derived class for polar Fourier representation for 2D images
    """

    def __init__(self, size, nrad=None, ntheta=None, dtype=np.float32):
        """
        Initialize an object for the polar Fourier transform class. `PolarFT` expects that
        images are real and uses only half of the `ntheta` values.

        :param size: The shape of the vectors for which to define the transform.
            May be a 2-tuple or an integer, in which case a square basis is assumed.
            Currently only square images are supported.
        :param nrad: The number of points in the radial dimension. Default is resolution // 2.
        :param ntheta: The number of points in the angular dimension. Default is 8 * nrad.
        :param dtype: dtype used to compute a polar frequency grid for evaluating the transform..
        """
        if isinstance(size, int):
            size = (size, size)
        ndim = len(size)
        assert ndim == 2, "Only two-dimensional grids are supported."
        assert len(set(size)) == 1, "Only square domains are supported."

        self.ndim = ndim
        self.sz = size
        self.nrad = nrad
        self.ntheta = ntheta
        self.dtype = dtype

        self._build()

        # this basis has complex coefficients
        self.coefficient_dtype = complex_type(self.dtype)

    def _build(self):
        """
        Build the internal data structure to 2D polar Fourier grid
        """
        logger.info("Represent 2D image in a polar Fourier grid")

        if self.nrad is None:
            self.nrad = self.sz[0] // 2

        if self.ntheta is None:
            # try to use the same number as Fast FB basis
            self.ntheta = 8 * self.nrad

        if self.ntheta % 2 == 1:
            msg = "Only even values for ntheta are supported."
            logger.error(msg)
            raise NotImplementedError(msg)

        self.count = self.nrad * (self.ntheta // 2)
        self._sz_prod = self.sz[0] * self.sz[1]

        # precompute the basis functions in 2D grids
        self.freqs = self._precomp()

    def _precomp(self):
        """
        Precompute the polar Fourier grid.
        """
        omega0 = 2 * np.pi / (2 * self.nrad - 1)
        dtheta = 2 * np.pi / self.ntheta

        # only need half size of ntheta
        freqs = np.zeros((2, self.ntheta // 2, self.nrad), dtype=self.dtype)
        for i in range(self.ntheta // 2):
            freqs[0, i] = np.cos(i * dtheta)
            freqs[1, i] = np.sin(i * dtheta)

        freqs *= omega0 * np.arange(self.nrad)

        return freqs.reshape(2, -1)

    def transform(self, x):
        """
        Evaluate coefficient in polar Fourier grid from those in standard 2D coordinate basis

        :param x: The Image instance representing coefficient array in the
            standard 2D coordinate basis to be evaluated.
        :return: The evaluation of the coefficient array `x` in the polar
            Fourier grid. This is an array of vectors whose first dimension
            corresponds to `x.shape[0]`, and last dimension equals `self.count`.
        """
        if x.dtype != self.dtype:
            raise TypeError(
                f"{self.__class__.__name__}.transform"
                f" Inconsistent dtypes x: {x.dtype} self: {self.dtype}"
            )

        if not isinstance(x, Image):
            raise TypeError(
                f"{self.__class__.__name__}.transform"
                f" passed numpy array instead of {Image}."
            )
        else:
            x = x.asnumpy()

        # Flatten stack
        stack_shape = x.shape[: -self.ndim]
        x = x.reshape(-1, *x.shape[-self.ndim :])

        # We expect the Image `x` to be real in order to take advantage of the conjugate
        # symmetry of the Fourier transform of a real valued image.
        if not np.isreal(x).all():
            raise TypeError(
                f"The Image `x` must be real valued. Found dtype {x.dtype}."
            )

        resolution = x.shape[-1]

        pf = nufft(x, self.freqs) / resolution**2

        return pf.reshape(*stack_shape, self.ntheta // 2, self.nrad)

    @staticmethod
    def half_to_full(pf):
        """
        Use the conjugate symmetry of pf to construct the full polar Fourier transform
        over all rays in [0, 360).

        :param pf: The precomputed half polar Fourier transform
            with shape (*stack_shape, ntheta//2, nrad)
        :return: The full polar Fourier transform with shape (*stack_shape, ntheta, nrad)
        """

        return np.concatenate((pf, np.conj(pf)), axis=-2)



import logging

from aspire.source.coordinates import BoxesCoordinateSource, CentersCoordinateSource
from aspire.source.image import (
    ArrayImageSource,
    ImageSource,
    IndexedSource,
    OrientedSource,
)
from aspire.source.relion import RelionSource
from aspire.source.simulation import Simulation

# isort: off
from aspire.source.micrograph import (
    ArrayMicrographSource,
    DiskMicrographSource,
    MicrographSimulation,
)

# isort: on
logger = logging.getLogger(__name__)



import logging

import click

from aspire.basis import FBBasis3D
from aspire.commands import log_level_option
from aspire.covariance import CovarianceEstimator
from aspire.noise import WhiteNoiseEstimator
from aspire.reconstruction import MeanEstimator
from aspire.source.relion import RelionSource
from aspire.utils.logging import setConsoleLoggingLevel

logger = logging.getLogger(__name__)


@click.command()
@click.option("--starfile", required=True, help="Path to starfile")
@click.option(
    "--data_folder", default=None, help="Path to mrcs files referenced in starfile"
)
@click.option(
    "--pixel_size", default=1, type=float, help="Pixel size of images in starfile"
)
@click.option(
    "--max_rows",
    default=None,
    type=int,
    help="Max. number of image rows to read from starfile",
)
@click.option(
    "--max_resolution",
    default=8,
    type=int,
    help="Resolution of downsampled images read from starfile",
)
@click.option("--cg_tol", default=1e-5, help="Tolerance for optimization convergence")
@log_level_option
def cov3d(
    starfile, data_folder, pixel_size, max_rows, max_resolution, cg_tol, loglevel
):
    """Estimate mean volume and covariance from a starfile."""

    # Set desired logging option for the command line
    setConsoleLoggingLevel(loglevel)

    source = RelionSource(
        starfile, data_folder=data_folder, pixel_size=pixel_size, max_rows=max_rows
    )

    source = source.downsample(max_resolution)
    source = source.cache()

    source = source.whiten()
    basis = FBBasis3D((max_resolution, max_resolution, max_resolution))
    mean_estimator = MeanEstimator(source, basis, batch_size=8192)
    mean_est = mean_estimator.estimate()

    noise_estimator = WhiteNoiseEstimator(source, batchSize=500)
    # Estimate the noise variance. This is needed for the covariance estimation step below.
    noise_variance = noise_estimator.estimate()
    logger.info(f"Noise Variance = {noise_variance}")

    # Passing in a mean_kernel argument to the following constructor speeds up some calculations
    covar_estimator = CovarianceEstimator(
        source, basis, mean_kernel=mean_estimator.kernel
    )
    covar_estimator.estimate(mean_est, noise_variance, tol=cg_tol)
